@@ -1,6 +1,6 @@
 /*
  * ProFTPD - mod_sftp packet IO
- * Copyright (c) 2008-2012 TJ Saunders
+ * Copyright (c) 2008-2016 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,15 +20,13 @@
  * give permission to link this program with OpenSSL, and distribute the
  * resulting executable, without including the source code for OpenSSL in the
  * source distribution.
- *
- * $Id: packet.h,v 1.10 2012-03-11 18:44:03 castaglia Exp $
  */
 
-#include "mod_sftp.h"
-
 #ifndef MOD_SFTP_PACKET_H
 #define MOD_SFTP_PACKET_H
 
+#include "mod_sftp.h"
+
 /* From RFC 4253, Section 6 */
 struct ssh2_packet {
   pool *pool;
@@ -114,4 +112,4 @@ int sftp_ssh2_packet_set_version(const char *);
 
 int sftp_ssh2_packet_set_client_alive(unsigned int, unsigned int);
 
-#endif
+#endif /* MOD_SFTP_PACKET_H */
