@@ -1,6 +1,6 @@
 /*
  * ProFTPD: mod_sql_odbc -- Support for connecting to databases via ODBC
- * Copyright (c) 2003-2017 TJ Saunders
+ * Copyright (c) 2003-2020 TJ Saunders
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -102,18 +102,23 @@ static conn_entry_t *sqlodbc_get_conn(char *name) {
 static void *sqlodbc_add_conn(pool *p, char *name, db_conn_t *conn) {
   conn_entry_t *entry = NULL;
 
-  if (!name || !conn || !p)
+  if (p == NULL ||
+      name == NULL ||
+      conn == NULL) {
+    errno = EINVAL;
     return NULL;
-  
-  if (sqlodbc_get_conn(name))
+  }
+
+  if (sqlodbc_get_conn(name) != NULL) {
+    errno = EEXIST;
     return NULL;
+  }
 
   entry = (conn_entry_t *) pcalloc(p, sizeof(conn_entry_t));
   entry->name = name;
   entry->data = conn;
 
   *((conn_entry_t **) push_array(conn_cache)) = entry;
-
   return entry;
 }
 
@@ -337,7 +342,7 @@ static modret_t *sqlodbc_get_error(cmd_rec *cmd, SQLSMALLINT handle_type,
   char numstr[20];
 
   memset(errstr, '\0', sizeof(errstr));
-  snprintf((char *) errstr, sizeof(errstr)-1, "%s", "(no data)");
+  pr_snprintf((char *) errstr, sizeof(errstr)-1, "%s", "(no data)");
 
   res = SQLGetDiagRec(handle_type, handle, recno++, state, &odbc_errno,
     errstr, sizeof(errstr), &errlen);
@@ -361,7 +366,7 @@ static modret_t *sqlodbc_get_error(cmd_rec *cmd, SQLSMALLINT handle_type,
    * have logged all the previous errors.
    */
   memset(numstr, '\0', sizeof(numstr));
-  snprintf(numstr, 20, "%d", (int) odbc_errno);
+  pr_snprintf(numstr, 20, "%d", (int) odbc_errno);
 
   return PR_ERROR_MSG(cmd, numstr, (char *) errstr);
 }
@@ -378,7 +383,8 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
   sd = (sql_data_t *) pcalloc(cmd->tmp_pool, sizeof(sql_data_t));
 
   res = SQLNumResultCols(conn->sth, &ncols);
-  if (res != SQL_SUCCESS && res != SQL_SUCCESS_WITH_INFO) {
+  if (res != SQL_SUCCESS &&
+      res != SQL_SUCCESS_WITH_INFO) {
     char *err;
     SQLCHAR *state = NULL;
 
@@ -406,7 +412,7 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
 
     switch (res) {
       case SQL_ERROR:
-        sql_log(DEBUG_WARN, "error fetching row %u: %s", sd->rnum + 1,
+        sql_log(DEBUG_WARN, "error fetching row %lu: %s", sd->rnum + 1,
           sqlodbc_errstr(SQL_HANDLE_STMT, conn->sth, NULL));
         return PR_ERROR(cmd);
 
@@ -416,7 +422,7 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
 
       case SQL_SUCCESS_WITH_INFO:
         /* Note: this deliberately falls through to the SQL_SUCCESS case. */
-        sql_log(DEBUG_WARN, "fetching row %u: %s", sd->rnum + 1,
+        sql_log(DEBUG_WARN, "fetching row %lu: %s", sd->rnum + 1,
           sqlodbc_errstr(SQL_HANDLE_STMT, conn->sth, NULL));
 
       case SQL_SUCCESS:
@@ -452,7 +458,7 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
                   if (SQLGetData(conn->sth, i, col_ctype, buf, col_size,
                       &buflen) != SQL_SUCCESS) {
                     sql_log(DEBUG_WARN, "error getting %s data for column %u, "
-                      "row %u: %s", sqlodbc_typestr(col_ctype), i,
+                      "row %lu: %s", sqlodbc_typestr(col_ctype), i,
                       sd->rnum + 1, sqlodbc_errstr(SQL_HANDLE_STMT, conn->sth,
                         NULL));
                     *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
@@ -461,14 +467,14 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
 
                   if (buflen == SQL_NO_TOTAL) {
                     sql_log(DEBUG_WARN, "notice: unable to determine total "
-                      "number of bytes remaining for %s column %u, row %u",
+                      "number of bytes remaining for %s column %u, row %lu",
                       sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                     *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                     break;
 
                   } else if (buflen == SQL_NULL_DATA) {
                     sql_log(DEBUG_WARN, "notice: data is NULL for %s column "
-                      "%u, row %u", sqlodbc_typestr(col_ctype), i,
+                      "%u, row %lu", sqlodbc_typestr(col_ctype), i,
                       sd->rnum + 1);
                     *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                     break;
@@ -498,7 +504,7 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
                 if (SQLGetData(conn->sth, i, col_ctype,
                     (SQLPOINTER) &col_cval, 0, &ind) != SQL_SUCCESS) {
                   sql_log(DEBUG_WARN, "error getting %s data for column %u, "
-                    "row %u: %s", sqlodbc_typestr(col_ctype), i,
+                    "row %lu: %s", sqlodbc_typestr(col_ctype), i,
                     sd->rnum + 1, sqlodbc_errstr(SQL_HANDLE_STMT, conn->sth,
                       NULL));
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
@@ -507,19 +513,19 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
 
                 if (ind == SQL_NO_TOTAL) {
                   sql_log(DEBUG_WARN, "notice: unable to determine total "
-                    "number of bytes remaining for %s column %u, row %u",
+                    "number of bytes remaining for %s column %u, row %lu",
                     sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
 
                 } else if (ind == SQL_NULL_DATA) {
                   sql_log(DEBUG_WARN, "notice: data is NULL for %s column %u, "
-                    "row %u", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
+                    "row %lu", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
                 }
 
-                snprintf(buf, sizeof(buf), "%hd", col_cval);
+                pr_snprintf(buf, sizeof(buf), "%hd", col_cval);
                 buf[sizeof(buf)-1] = '\0';
 
                 *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, buf);
@@ -536,7 +542,7 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
                 if (SQLGetData(conn->sth, i, col_ctype,
                     (SQLPOINTER) &col_cval, 0, &ind) != SQL_SUCCESS) {
                   sql_log(DEBUG_WARN, "error getting %s data for column %u, "
-                    "row %u: %s", sqlodbc_typestr(col_ctype), i,
+                    "row %lu: %s", sqlodbc_typestr(col_ctype), i,
                     sd->rnum + 1, sqlodbc_errstr(SQL_HANDLE_STMT, conn->sth,
                       NULL));
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
@@ -545,19 +551,19 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
 
                 if (ind == SQL_NO_TOTAL) {
                   sql_log(DEBUG_WARN, "notice: unable to determine total "
-                    "number of bytes remaining for %s column %u, row %u",
+                    "number of bytes remaining for %s column %u, row %lu",
                     sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
 
                 } else if (ind == SQL_NULL_DATA) {
                   sql_log(DEBUG_WARN, "notice: data is NULL for %s column %u, "
-                    "row %u", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
+                    "row %lu", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
                 }
 
-                snprintf(buf, sizeof(buf), "%hd", col_cval);
+                pr_snprintf(buf, sizeof(buf), "%hd", col_cval);
                 buf[sizeof(buf)-1] = '\0';
 
                 *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, buf);
@@ -574,7 +580,7 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
                 if (SQLGetData(conn->sth, i, col_ctype,
                     (SQLPOINTER) &col_cval, 0, &ind) != SQL_SUCCESS) {
                   sql_log(DEBUG_WARN, "error getting %s data for column %u, "
-                    "row %u: %s", sqlodbc_typestr(col_ctype), i,
+                    "row %lu: %s", sqlodbc_typestr(col_ctype), i,
                     sd->rnum + 1, sqlodbc_errstr(SQL_HANDLE_STMT, conn->sth,
                       NULL));
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
@@ -583,19 +589,19 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
 
                 if (ind == SQL_NO_TOTAL) {
                   sql_log(DEBUG_WARN, "notice: unable to determine total "
-                    "number of bytes remaining for %s column %u, row %u",
+                    "number of bytes remaining for %s column %u, row %lu",
                     sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
 
                 } else if (ind == SQL_NULL_DATA) {
                   sql_log(DEBUG_WARN, "notice: data is NULL for %s column %u, "
-                    "row %u", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
+                    "row %lu", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
                 }
 
-                snprintf(buf, sizeof(buf), "%d", col_cval);
+                pr_snprintf(buf, sizeof(buf), "%d", col_cval);
                 buf[sizeof(buf)-1] = '\0';
 
                 *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, buf);
@@ -612,7 +618,7 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
                 if (SQLGetData(conn->sth, i, col_ctype,
                     (SQLPOINTER) &col_cval, 0, &ind) != SQL_SUCCESS) {
                   sql_log(DEBUG_WARN, "error getting %s data for column %u, "
-                    "row %u: %s", sqlodbc_typestr(col_ctype), i,
+                    "row %lu: %s", sqlodbc_typestr(col_ctype), i,
                     sd->rnum + 1, sqlodbc_errstr(SQL_HANDLE_STMT, conn->sth,
                       NULL));
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
@@ -621,19 +627,19 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
 
                 if (ind == SQL_NO_TOTAL) {
                   sql_log(DEBUG_WARN, "notice: unable to determine total "
-                    "number of bytes remaining for %s column %u, row %u",
+                    "number of bytes remaining for %s column %u, row %lu",
                     sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
 
                 } else if (ind == SQL_NULL_DATA) {
                   sql_log(DEBUG_WARN, "notice: data is NULL for %s column %u, "
-                    "row %u", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
+                    "row %lu", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
                 }
 
-                snprintf(buf, sizeof(buf), "%ld", col_cval);
+                pr_snprintf(buf, sizeof(buf), "%ld", col_cval);
                 buf[sizeof(buf)-1] = '\0';
 
                 *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, buf);
@@ -650,7 +656,7 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
                 if (SQLGetData(conn->sth, i, col_ctype,
                     (SQLPOINTER) &col_cval, 0, &ind) != SQL_SUCCESS) {
                   sql_log(DEBUG_WARN, "error getting %s data for column %u, "
-                    "row %u: %s", sqlodbc_typestr(col_ctype), i,
+                    "row %lu: %s", sqlodbc_typestr(col_ctype), i,
                     sd->rnum + 1, sqlodbc_errstr(SQL_HANDLE_STMT, conn->sth,
                       NULL));
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
@@ -659,19 +665,19 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
 
                 if (ind == SQL_NO_TOTAL) {
                   sql_log(DEBUG_WARN, "notice: unable to determine total "
-                    "number of bytes remaining for %s column %u, row %u",
+                    "number of bytes remaining for %s column %u, row %lu",
                     sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
 
                 } else if (ind == SQL_NULL_DATA) {
                   sql_log(DEBUG_WARN, "notice: data is NULL for %s column %u, "
-                    "row %u", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
+                    "row %lu", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
                 }
 
-                snprintf(buf, sizeof(buf), "%ld", col_cval);
+                pr_snprintf(buf, sizeof(buf), "%ld", col_cval);
                 buf[sizeof(buf)-1] = '\0';
 
                 *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, buf);
@@ -688,7 +694,7 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
                 if (SQLGetData(conn->sth, i, col_ctype,
                     (SQLPOINTER) &col_cval, 0, &ind) != SQL_SUCCESS) {
                   sql_log(DEBUG_WARN, "error getting %s data for column %u, "
-                    "row %u: %s", sqlodbc_typestr(col_ctype), i,
+                    "row %lu: %s", sqlodbc_typestr(col_ctype), i,
                     sd->rnum + 1, sqlodbc_errstr(SQL_HANDLE_STMT, conn->sth,
                       NULL));
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
@@ -697,19 +703,19 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
 
                 if (ind == SQL_NO_TOTAL) {
                   sql_log(DEBUG_WARN, "notice: unable to determine total "
-                    "number of bytes remaining for %s column %u, row %u",
+                    "number of bytes remaining for %s column %u, row %lu",
                     sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
 
                 } else if (ind == SQL_NULL_DATA) {
                   sql_log(DEBUG_WARN, "notice: data is NULL for %s column %u, "
-                    "row %u", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
+                    "row %lu", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
                 }
 
-                snprintf(buf, sizeof(buf), "%f", col_cval);
+                pr_snprintf(buf, sizeof(buf), "%f", col_cval);
                 buf[sizeof(buf)-1] = '\0';
 
                 *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, buf);
@@ -726,7 +732,7 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
                 if (SQLGetData(conn->sth, i, col_ctype,
                     (SQLPOINTER) &col_cval, 0, &ind) != SQL_SUCCESS) {
                   sql_log(DEBUG_WARN, "error getting %s data for column %u, "
-                    "row %u: %s", sqlodbc_typestr(col_ctype), i,
+                    "row %lu: %s", sqlodbc_typestr(col_ctype), i,
                     sd->rnum + 1, sqlodbc_errstr(SQL_HANDLE_STMT, conn->sth,
                       NULL));
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
@@ -735,19 +741,19 @@ static modret_t *sqlodbc_get_data(cmd_rec *cmd, db_conn_t *conn) {
 
                 if (ind == SQL_NO_TOTAL) {
                   sql_log(DEBUG_WARN, "notice: unable to determine total "
-                    "number of bytes remaining for %s column %u, row %u",
+                    "number of bytes remaining for %s column %u, row %lu",
                     sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
 
                 } else if (ind == SQL_NULL_DATA) {
                   sql_log(DEBUG_WARN, "notice: data is NULL for %s column %u, "
-                    "row %u", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
+                    "row %lu", sqlodbc_typestr(col_ctype), i, sd->rnum + 1);
                   *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, "");
                   break;
                 }
 
-                snprintf(buf, sizeof(buf), "%f", col_cval);
+                pr_snprintf(buf, sizeof(buf), "%f", col_cval);
                 buf[sizeof(buf)-1] = '\0';
 
                 *((char **) push_array(dh)) = pstrdup(cmd->tmp_pool, buf);
@@ -801,7 +807,8 @@ MODRET sqlodbc_open(cmd_rec *cmd) {
   entry = sqlodbc_get_conn(cmd->argv[0]);
   if (entry == NULL) {
     sql_log(DEBUG_FUNC, "%s", "exiting \todbc cmd_open");
-    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION,
+      pstrcat(cmd->tmp_pool, "unknown named connection: ", cmd->argv[0], NULL));
   } 
 
   conn = (db_conn_t *) entry->data;
@@ -856,7 +863,8 @@ MODRET sqlodbc_open(cmd_rec *cmd) {
 
   res = SQLConnect(conn->dbh, (SQLCHAR *) conn->dsn, SQL_NTS,
     (SQLCHAR *) conn->user, SQL_NTS, (SQLCHAR *) conn->pass, SQL_NTS);
-  if (res != SQL_SUCCESS) {
+  if (res != SQL_SUCCESS &&
+      res != SQL_SUCCESS_WITH_INFO) {
     sql_log(DEBUG_FUNC, "error connecting to dsn '%s': %s", conn->dsn,
       sqlodbc_strerror(res));
     sql_log(DEBUG_FUNC, "%s", "exiting \todbc cmd_open");
@@ -1009,7 +1017,8 @@ MODRET sqlodbc_close(cmd_rec *cmd) {
   entry = sqlodbc_get_conn(cmd->argv[0]);
   if (entry == NULL) {
     sql_log(DEBUG_FUNC, "%s", "exiting \todbc cmd_close");
-    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION,
+      pstrcat(cmd->tmp_pool, "unknown named connection: ", cmd->argv[0], NULL));
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1092,10 +1101,23 @@ MODRET sqlodbc_def_conn(cmd_rec *cmd) {
 
   /* Insert the new conn_info into the connection hash */
   entry = sqlodbc_add_conn(conn_pool, name, (void *) conn);
+  if (entry == NULL &&
+      errno == EEXIST) {
+    /* Log only connections named other than "default", for debugging
+     * misconfigurations with multiple different SQLNamedConnectInfo
+     * directives using the same name.
+     */
+    if (strcmp(name, "default") != 0) {
+      sql_log(DEBUG_FUNC, "named connection '%s' already exists", name);
+    }
+
+    entry = sqlodbc_get_conn(name);
+  }
+
   if (entry == NULL) {
     sql_log(DEBUG_FUNC, "%s", "exiting \todbc cmd_defineconnection");
     return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION,
-      "named connection already exists");
+      "error adding named connection");
   }
 
   if (cmd->argc >= 5) {
@@ -1137,7 +1159,8 @@ MODRET sqlodbc_select(cmd_rec *cmd) {
   entry = sqlodbc_get_conn(cmd->argv[0]);
   if (entry == NULL) {
     sql_log(DEBUG_FUNC, "%s", "exiting \todbc cmd_select");
-    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION,
+      pstrcat(cmd->tmp_pool, "unknown named connection: ", cmd->argv[0], NULL));
   }
  
   conn = (db_conn_t *) entry->data;
@@ -1286,7 +1309,8 @@ MODRET sqlodbc_insert(cmd_rec *cmd) {
   entry = sqlodbc_get_conn(cmd->argv[0]);
   if (entry == NULL) {
     sql_log(DEBUG_FUNC, "%s", "exiting \todbc cmd_insert");
-    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION,
+      pstrcat(cmd->tmp_pool, "unknown named connection: ", cmd->argv[0], NULL));
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1377,7 +1401,8 @@ MODRET sqlodbc_update(cmd_rec *cmd) {
   entry = sqlodbc_get_conn(cmd->argv[0]);
   if (entry == NULL) {
     sql_log(DEBUG_FUNC, "%s", "exiting \todbc cmd_update");
-    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION,
+      pstrcat(cmd->tmp_pool, "unknown named connection: ", cmd->argv[0], NULL));
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1484,7 +1509,8 @@ MODRET sqlodbc_query(cmd_rec *cmd) {
   entry = sqlodbc_get_conn(cmd->argv[0]);
   if (entry == NULL) {
     sql_log(DEBUG_FUNC, "%s", "exiting \todbc cmd_query");
-    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION,
+      pstrcat(cmd->tmp_pool, "unknown named connection: ", cmd->argv[0], NULL));
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1587,7 +1613,8 @@ MODRET sqlodbc_quote(cmd_rec *cmd) {
   entry = sqlodbc_get_conn(cmd->argv[0]);
   if (entry == NULL) {
     sql_log(DEBUG_FUNC, "%s", "exiting \todbc cmd_escapestring");
-    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION,
+      pstrcat(cmd->tmp_pool, "unknown named connection: ", cmd->argv[0], NULL));
   }
 
   /* Make sure the connection is open. */
@@ -1630,22 +1657,6 @@ MODRET sqlodbc_exit(cmd_rec *cmd) {
   return PR_HANDLED(cmd);
 }
 
-MODRET sqlodbc_checkauth(cmd_rec *cmd) {
-  sql_log(DEBUG_FUNC, "%s", "entering \todbc cmd_checkauth");
-
-  if (cmd->argc != 3) {
-    sql_log(DEBUG_FUNC, "exiting \todbc cmd_checkauth");
-    return PR_ERROR_MSG(cmd, MOD_SQL_ODBC_VERSION, "badly formed request");
-  }
-
-  /* This mod_sql backend doesn't support any database-specific password
-   * checking mechanisms.
-   */
-
-  sql_log(DEBUG_FUNC, "%s", "exiting \todbc cmd_checkauth");
-  return PR_ERROR(cmd);
-}
-
 MODRET sqlodbc_identify(cmd_rec *cmd) {
   sql_data_t *sd = NULL;
 
@@ -1673,7 +1684,6 @@ static cmdtable sqlodbc_cmdtable[] = {
   { CMD, "sql_query",		G_NONE, sqlodbc_query,	FALSE, FALSE },
   { CMD, "sql_escapestring",	G_NONE, sqlodbc_quote,	FALSE, FALSE },
   { CMD, "sql_exit",		G_NONE, sqlodbc_exit,	FALSE, FALSE },
-  { CMD, "sql_checkauth",	G_NONE, sqlodbc_checkauth, FALSE, FALSE },
   { CMD, "sql_identify",	G_NONE, sqlodbc_identify, FALSE, FALSE },
 
   { 0, NULL }
@@ -1767,13 +1777,16 @@ static int sqlodbc_init(void) {
 static int sqlodbc_sess_init(void) {
   config_rec *c;
 
-  if (conn_pool == NULL) {
-    conn_pool = make_sub_pool(session.pool);
-    pr_pool_tag(conn_pool, "ODBC connection pool");
+  if (conn_pool != NULL) {
+    destroy_pool(conn_pool);
+    conn_cache = NULL;
   }
 
+  conn_pool = make_sub_pool(session.pool);
+  pr_pool_tag(conn_pool, "ODBC connection pool");
+
   if (conn_cache == NULL) {
-    conn_cache = make_array(make_sub_pool(session.pool), DEF_CONN_POOL_SIZE,
+    conn_cache = make_array(conn_pool, DEF_CONN_POOL_SIZE,
       sizeof(conn_entry_t *));
   }
 
