@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2013 The ProFTPD Project team
+ * Copyright (c) 2001-2016 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,9 +24,7 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/* BSD socket manipulation tools.
- * $Id: inet.h,v 1.40 2013-02-07 15:44:29 castaglia Exp $
- */
+/* BSD socket manipulation tools. */
 
 #ifndef PR_INET_H
 #define PR_INET_H
@@ -112,7 +110,7 @@ typedef struct conn_struc {
   pr_netio_stream_t *instrm, *outstrm;	/* Input/Output streams */
 
   /* Remote address of the connection. */
-  pr_netaddr_t *remote_addr;
+  const pr_netaddr_t *remote_addr;
 
   /* Remote port of the connection. */
   int remote_port;
@@ -121,7 +119,7 @@ typedef struct conn_struc {
   const char *remote_name;
 
   /* Local address of the connection. */
-  pr_netaddr_t *local_addr;
+  const pr_netaddr_t *local_addr;
 
   /* Local port of the connection. */
   int local_port;
@@ -131,7 +129,7 @@ typedef struct conn_struc {
 /* Used for event data for events related to opening of sockets */
 struct socket_ctx {
   server_rec *server;
-  pr_netaddr_t *addr;
+  const pr_netaddr_t *addr;
   int sockfd;
 };
 
@@ -140,9 +138,9 @@ void pr_inet_clear(void);
 int pr_inet_reverse_dns(pool *, int);
 int pr_inet_getservport(pool *, const char *, const char *);
 pr_netaddr_t *pr_inet_getaddr(pool *, const char *, array_header **);
-conn_t *pr_inet_copy_conn(pool *, conn_t*);
-conn_t *pr_inet_create_conn(pool *, int, pr_netaddr_t *, int, int);
-conn_t *pr_inet_create_conn_portrange(pool *, pr_netaddr_t *, int, int);
+conn_t *pr_inet_copy_conn(pool *, conn_t *);
+conn_t *pr_inet_create_conn(pool *, int, const pr_netaddr_t *, int, int);
+conn_t *pr_inet_create_conn_portrange(pool *, const pr_netaddr_t *, int, int);
 void pr_inet_close(pool *, conn_t *);
 void pr_inet_lingering_abort(pool *, conn_t *, long);
 void pr_inet_lingering_close(pool *, conn_t *, long);
@@ -160,14 +158,14 @@ int pr_inet_listen(pool *p, conn_t *conn, int backlog, int flags);
 
 int pr_inet_resetlisten(pool *, conn_t *);
 int pr_inet_accept_nowait(pool *, conn_t *);
-int pr_inet_connect(pool *, conn_t *, pr_netaddr_t *, int);
-int pr_inet_connect_nowait(pool *, conn_t *, pr_netaddr_t *, int);
+int pr_inet_connect(pool *, conn_t *, const pr_netaddr_t *, int);
+int pr_inet_connect_nowait(pool *, conn_t *, const pr_netaddr_t *, int);
 int pr_inet_get_conn_info(conn_t *, int);
 conn_t *pr_inet_accept(pool *, conn_t *, conn_t *, int, int, unsigned char);
-conn_t *pr_inet_openrw(pool *, conn_t *, pr_netaddr_t *, int, int, int,
+conn_t *pr_inet_openrw(pool *, conn_t *, const pr_netaddr_t *, int, int, int,
   int, int);
-int pr_inet_generate_socket_event(const char *, server_rec *, pr_netaddr_t *,
-  int);
+int pr_inet_generate_socket_event(const char *, server_rec *,
+  const pr_netaddr_t *, int);
 
 void init_inet(void);
 
