@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2016 The ProFTPD Project
+ * Copyright (c) 2001-2020 The ProFTPD Project
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -55,7 +55,7 @@ static int sendline(int flags, char *fmt, ...)
 #define LS_FL_NO_ERROR_IF_ABSENT	0x0001
 #define LS_FL_LIST_ONLY			0x0002
 #define LS_FL_NLST_ONLY			0x0004
-#define LS_FL_NO_ADJUSTED_SYMLINKS	0x0008
+#define LS_FL_ADJUSTED_SYMLINKS		0x0008
 #define LS_FL_SORTED_NLST		0x0010
 static unsigned long list_flags = 0UL;
 
@@ -205,9 +205,8 @@ static void push_cwd(char *_cwd, unsigned char *symhold) {
   if (!_cwd)
     _cwd = cwd;
 
-  *symhold = show_symlinks_hold;
   sstrncpy(_cwd, pr_fs_getcwd(), PR_TUNABLE_PATH_MAX + 1);
-  *symhold = list_show_symlinks;
+  *symhold = show_symlinks_hold = list_show_symlinks;
 }
 
 static void pop_cwd(char *_cwd, unsigned char *symhold) {
@@ -374,7 +373,7 @@ static int sendline(int flags, char *fmt, ...) {
   }
 
   va_start(msg, fmt);
-  vsnprintf(buf, sizeof(buf), fmt, msg);
+  pr_vsnprintf(buf, sizeof(buf), fmt, msg);
   va_end(msg);
 
   buf[sizeof(buf)-1] = '\0';
@@ -420,13 +419,12 @@ static int sendline(int flags, char *fmt, ...) {
 }
 
 static void ls_done(cmd_rec *cmd) {
-  int quiet = FALSE;
+  pr_data_close2();
 
-  if (session.sf_flags & SF_ABORT) {
-    quiet = TRUE;
+  if (!(session.sf_flags & SF_ABORT)) {
+    /* If the transfer was not aborted, consider it a success. */
+    pr_response_add(R_226, _("Transfer complete"));
   }
-
-  pr_data_close(quiet);
 }
 
 static char units[6][2] = 
@@ -434,7 +432,7 @@ static char units[6][2] =
 
 static void ls_fmt_filesize(char *buf, size_t buflen, off_t sz) {
   if (!opt_h || sz < 1000) {
-    snprintf(buf, buflen, "%8" PR_LU, (pr_off_t) sz);
+    pr_snprintf(buf, buflen, "%8" PR_LU, (pr_off_t) sz);
 
   } else {
     register unsigned int i = 0;
@@ -446,7 +444,7 @@ static void ls_fmt_filesize(char *buf, size_t buflen, off_t sz) {
       i++;
     }
 
-    snprintf(buf, buflen, "%7.1f%s", size, units[i]);
+    pr_snprintf(buf, buflen, "%7.1f%s", size, units[i]);
   }
 }
 
@@ -537,7 +535,7 @@ static int listfile(cmd_rec *cmd, pool *p, const char *resp_code,
           char replace[32];
 
           memset(replace, '\0', sizeof(replace));
-          replace_len = snprintf(replace, sizeof(replace)-1, "\\%03o",
+          replace_len = pr_snprintf(replace, sizeof(replace)-1, "\\%03o",
             display_name[i]);
 
           for (k = 0; k < replace_len; k++) {
@@ -572,12 +570,12 @@ static int listfile(cmd_rec *cmd, pool *p, const char *resp_code,
           return 0;
         }
 
-        if (list_flags & LS_FL_NO_ADJUSTED_SYMLINKS) {
-          len = pr_fsio_readlink(name, m, sizeof(m) - 1);
-
-        } else {
+        if (list_flags & LS_FL_ADJUSTED_SYMLINKS) {
           len = dir_readlink(p, name, m, sizeof(m) - 1,
             PR_DIR_READLINK_FL_HANDLE_REL_PATH);
+
+        } else {
+          len = pr_fsio_readlink(name, m, sizeof(m) - 1);
         }
 
         if (len < 0) {
@@ -615,12 +613,12 @@ static int listfile(cmd_rec *cmd, pool *p, const char *resp_code,
         return 0;
       }
 
-      if (list_flags & LS_FL_NO_ADJUSTED_SYMLINKS) {
-        len = pr_fsio_readlink(name, l, sizeof(l) - 1);
-
-      } else {
+      if (list_flags & LS_FL_ADJUSTED_SYMLINKS) {
         len = dir_readlink(p, name, l, sizeof(l) - 1,
           PR_DIR_READLINK_FL_HANDLE_REL_PATH);
+
+      } else {
+        len = pr_fsio_readlink(name, l, sizeof(l) - 1);
       }
 
       if (len < 0) {
@@ -764,10 +762,10 @@ static int listfile(cmd_rec *cmd, pool *p, const char *resp_code,
         m[1] = (mode & S_IRUSR) ? 'r' : '-';
 
         if (ls_curtime - sort_time > 180 * 24 * 60 * 60) {
-          snprintf(timeline, sizeof(timeline), "%5d", t->tm_year+1900);
+          pr_snprintf(timeline, sizeof(timeline), "%5d", t->tm_year+1900);
 
         } else {
-          snprintf(timeline, sizeof(timeline), "%02d:%02d", t->tm_hour,
+          pr_snprintf(timeline, sizeof(timeline), "%02d:%02d", t->tm_hour,
             t->tm_min);
         }
 
@@ -775,13 +773,13 @@ static int listfile(cmd_rec *cmd, pool *p, const char *resp_code,
 
         if (opt_1) {
           /* One file per line, with no info other than the file name.  Easy. */
-          snprintf(nameline, sizeof(nameline)-1, "%s",
+          pr_snprintf(nameline, sizeof(nameline)-1, "%s",
             pr_fs_encode_path(cmd->tmp_pool, display_name));
 
         } else {
           if (!opt_n) {
             /* Format nameline using user/group names. */
-            snprintf(nameline, sizeof(nameline)-1,
+            pr_snprintf(nameline, sizeof(nameline)-1,
               "%s %3d %-8s %-8s %s %s %2d %s %s", m, (int) st.st_nlink,
               MAP_UID(st.st_uid), MAP_GID(st.st_gid), s,
               months[t->tm_mon], t->tm_mday, timeline,
@@ -789,7 +787,7 @@ static int listfile(cmd_rec *cmd, pool *p, const char *resp_code,
 
           } else {
             /* Format nameline using user/group IDs. */
-            snprintf(nameline, sizeof(nameline)-1,
+            pr_snprintf(nameline, sizeof(nameline)-1,
               "%s %3d %-8u %-8u %s %s %2d %s %s", m, (int) st.st_nlink,
               (unsigned) st.st_uid, (unsigned) st.st_gid, s,
               months[t->tm_mon], t->tm_mday, timeline,
@@ -819,7 +817,7 @@ static int listfile(cmd_rec *cmd, pool *p, const char *resp_code,
 
           if (!opt_L && list_show_symlinks) {
             if (sizeof(nameline) - strlen(nameline) > 4) {
-              snprintf(buf, sizeof(nameline) - strlen(nameline) - 4,
+              pr_snprintf(buf, sizeof(nameline) - strlen(nameline) - 4,
                 " -> %s", l);
             } else {
               pr_log_pri(PR_LOG_NOTICE, "notice: symlink '%s' yields an "
@@ -921,7 +919,7 @@ static void addfile(cmd_rec *cmd, const char *name, const char *suffix,
 
   p = (struct filename *) pcalloc(fpool, sizeof(struct filename));
   p->line = pcalloc(fpool, l + 2);
-  snprintf(p->line, l + 1, "%s%s", name, suffix);
+  pr_snprintf(p->line, l + 1, "%s%s", name, suffix);
 
   if (tail) {
     tail->down = p;
@@ -1529,7 +1527,7 @@ static void ls_terminate(void) {
     discard_output();
 
     if (!XFER_ABORTED) {
-      /* An error has occured, other than client ABOR */
+      /* An error has occurred, other than client ABOR */
       if (ls_errno) {
         pr_data_abort(ls_errno,FALSE);
 
@@ -1946,7 +1944,7 @@ static int dolist(cmd_rec *cmd, const char *opt, const char *resp_code,
 
       pw = pr_auth_getpwnam(cmd->tmp_pool, i ? pbuffer : session.user);
       if (pw) {
-        snprintf(pbuffer, sizeof(pbuffer), "%s%s", pw->pw_dir, p);
+        pr_snprintf(pbuffer, sizeof(pbuffer), "%s%s", pw->pw_dir, p);
 
       } else {
         pbuffer[0] = '\0';
@@ -1960,9 +1958,6 @@ static int dolist(cmd_rec *cmd, const char *opt, const char *resp_code,
       if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
         int xerrno = errno;
 
-        pr_response_add_err(R_450, "%s: %s", (char *) cmd->argv[0],
-          strerror(xerrno));
-
         pr_cmd_set_errno(cmd, xerrno);
         errno = xerrno;
         return -1;
@@ -2131,7 +2126,7 @@ static int dolist(cmd_rec *cmd, const char *opt, const char *resp_code,
           /* Recurse into the directory. */
           push_cwd(cwd_buf, &symhold);
 
-          if (!pr_fsio_chdir_canon(*path, !opt_L && list_show_symlinks)) {
+          if (pr_fsio_chdir_canon(*path, !opt_L && list_show_symlinks) == 0) {
             int res = 0;
 
             list_ndepth.curr++;
@@ -2150,6 +2145,9 @@ static int dolist(cmd_rec *cmd, const char *opt, const char *resp_code,
               }
               return -1;
             }
+
+          } else {
+            pop_cwd(cwd_buf, &symhold);
           }
         }
 
@@ -2203,9 +2201,6 @@ static int dolist(cmd_rec *cmd, const char *opt, const char *resp_code,
       if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
         int xerrno = errno;
 
-        pr_response_add_err(R_450, "%s: %s", (char *) cmd->argv[0], 
-          strerror(xerrno));
-
         pr_cmd_set_errno(cmd, xerrno);
         errno = xerrno;
         return -1;
@@ -2289,7 +2284,7 @@ static int nlstfile(cmd_rec *cmd, const char *file) {
         char replace[32];
 
         memset(replace, '\0', sizeof(replace));
-        replace_len = snprintf(replace, sizeof(replace)-1, "\\%03o",
+        replace_len = pr_snprintf(replace, sizeof(replace)-1, "\\%03o",
           display_name[i]);
 
         for (k = 0; k < replace_len; k++) {
@@ -2441,12 +2436,12 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
       }
     }
 
-    if (list_flags & LS_FL_NO_ADJUSTED_SYMLINKS) {
-      i = pr_fsio_readlink(p, file, sizeof(file) - 1);
-
-    } else {
+    if (list_flags & LS_FL_ADJUSTED_SYMLINKS) {
       i = dir_readlink(cmd->tmp_pool, p, file, sizeof(file) - 1,
         PR_DIR_READLINK_FL_HANDLE_REL_PATH);
+
+    } else {
+      i = pr_fsio_readlink(p, file, sizeof(file) - 1);
     }
 
     if (i > 0) {
@@ -3060,8 +3055,8 @@ MODRET ls_nlst(cmd_rec *cmd) {
     pb[i] = '\0';
 
     pw = pr_auth_getpwnam(cmd->tmp_pool, i ? pb : session.user);
-    if (pw) {
-      snprintf(pb, sizeof(pb), "%s%s", pw->pw_dir, p);
+    if (pw != NULL) {
+      pr_snprintf(pb, sizeof(pb), "%s%s", pw->pw_dir, p);
       sstrncpy(buf, pb, sizeof(buf));
       target = buf;
     }
@@ -3104,9 +3099,6 @@ MODRET ls_nlst(cmd_rec *cmd) {
             if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
               int xerrno = errno;
 
-              pr_response_add_err(R_450, "%s: %s", (char *) cmd->argv[0], 
-                strerror(xerrno));
-
               pr_cmd_set_errno(cmd, xerrno);
               errno = xerrno;
               return PR_ERROR(cmd);
@@ -3131,9 +3123,6 @@ MODRET ls_nlst(cmd_rec *cmd) {
           if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
             int xerrno = errno;
 
-            pr_response_add_err(R_450, "%s: %s", (char *) cmd->argv[0],
-              strerror(xerrno));
-
             pr_cmd_set_errno(cmd, xerrno);
             errno = xerrno;
             return PR_ERROR(cmd);
@@ -3157,9 +3146,6 @@ MODRET ls_nlst(cmd_rec *cmd) {
     if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
       int xerrno = errno;
 
-      pr_response_add_err(R_450, "%s: %s", (char *) cmd->argv[0],
-        strerror(xerrno));
-
       pr_cmd_set_errno(cmd, xerrno);
       errno = xerrno;
       return PR_ERROR(cmd);
@@ -3250,9 +3236,6 @@ MODRET ls_nlst(cmd_rec *cmd) {
         if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
           xerrno = errno;
 
-          pr_response_add_err(R_450, "%s: %s", (char *) cmd->argv[0],
-            strerror(xerrno));
-
           pr_cmd_set_errno(cmd, xerrno);
           errno = xerrno;
           return PR_ERROR(cmd);
@@ -3287,9 +3270,6 @@ MODRET ls_nlst(cmd_rec *cmd) {
             if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
               xerrno = errno;
 
-              pr_response_add_err(R_450, "%s: %s", (char *) cmd->argv[0],
-                strerror(xerrno));
-
               pr_cmd_set_errno(cmd, xerrno);
               errno = xerrno;
               return PR_ERROR(cmd);
@@ -3328,9 +3308,6 @@ MODRET ls_nlst(cmd_rec *cmd) {
         if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
           xerrno = errno;
 
-          pr_response_add_err(R_450, "%s: %s", (char *) cmd->argv[0],
-            strerror(xerrno));
-
           pr_cmd_set_errno(cmd, xerrno);
           errno = xerrno;
           return PR_ERROR(cmd);
@@ -3353,9 +3330,6 @@ MODRET ls_nlst(cmd_rec *cmd) {
       if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
         int xerrno = errno;
 
-        pr_response_add_err(R_450, "%s: %s", (char *) cmd->argv[0],
-          strerror(xerrno));
-
         pr_cmd_set_errno(cmd, xerrno);
         errno = xerrno;
         return PR_ERROR(cmd);
@@ -3564,8 +3538,11 @@ MODRET set_listoptions(cmd_rec *cmd) {
       } else if (strcasecmp(cmd->argv[i], "NoErrorIfAbsent") == 0) {
         flags |= LS_FL_NO_ERROR_IF_ABSENT;
 
+      } else if (strcasecmp(cmd->argv[i], "AdjustedSymlinks") == 0) {
+        flags |= LS_FL_ADJUSTED_SYMLINKS;
+
       } else if (strcasecmp(cmd->argv[i], "NoAdjustedSymlinks") == 0) {
-        flags |= LS_FL_NO_ADJUSTED_SYMLINKS;
+        /* Ignored, for backward compatibility. */
 
       } else if (strcasecmp(cmd->argv[i], "SortedNLST") == 0) {
         flags |= LS_FL_SORTED_NLST;
