@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001-2016 The ProFTPD Project team
+ * Copyright (c) 2001-2020 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -584,7 +584,7 @@ server_rec *pr_ipbind_get_server(const pr_netaddr_t *addr, unsigned int port) {
   unsigned int match_count = 0;
 
   /* If we've got a binding configured for this exact address, return it
-   * straightaway.
+   * straight away.
    */
   ipbind = pr_ipbind_find(addr, port, TRUE);
   if (ipbind != NULL) {
@@ -799,7 +799,8 @@ int pr_ipbind_open(const pr_netaddr_t *addr, unsigned int port,
     for (i = 0; i < ipbind->ib_namebinds->nelts; i++) {
       pr_namebind_t *nb = namebinds[i];
 
-      res = pr_namebind_open(nb->nb_name, nb->nb_server->addr);
+      res = pr_namebind_open(nb->nb_name, nb->nb_server->addr,
+        nb->nb_server_port);
       if (res < 0) {
         pr_trace_msg(trace_channel, 2,
           "notice: unable to open namebind '%s': %s", nb->nb_name,
@@ -894,7 +895,7 @@ int pr_namebind_create(server_rec *server, const char *name,
   }
 
   /* Make sure we can add this namebind. */
-  if (!ipbind->ib_namebinds) {
+  if (ipbind->ib_namebinds == NULL) {
     ipbind->ib_namebinds = make_array(binding_pool, 0, sizeof(pr_namebind_t *));
 
   } else {
@@ -925,6 +926,7 @@ int pr_namebind_create(server_rec *server, const char *name,
   namebind = (pr_namebind_t *) pcalloc(server->pool, sizeof(pr_namebind_t));
   namebind->nb_name = name;
   namebind->nb_server = server;
+  namebind->nb_server_port = server_port;
   namebind->nb_isactive = FALSE;
 
   if (pr_str_is_fnmatch(name) == TRUE) {
@@ -1061,7 +1063,8 @@ pr_namebind_t *pr_namebind_find(const char *name, const pr_netaddr_t *addr,
        */
       if (namebind->nb_name != NULL) {
         pr_trace_msg(trace_channel, 17,
-          "namebind #%u: %s", i, namebind->nb_name);
+          "namebind #%u: %s (%s)", i, namebind->nb_name,
+          namebind->nb_isactive ? "active" : "inactive");
 
         if (namebind->nb_iswildcard == FALSE) {
           if (strcasecmp(namebind->nb_name, name) == 0) {
@@ -1120,7 +1123,8 @@ unsigned int pr_namebind_count(server_rec *srv) {
   return count;
 }
 
-int pr_namebind_open(const char *name, const pr_netaddr_t *addr) {
+int pr_namebind_open(const char *name, const pr_netaddr_t *addr,
+    unsigned int server_port) {
   pr_namebind_t *namebind = NULL;
   unsigned int port;
 
@@ -1131,6 +1135,9 @@ int pr_namebind_open(const char *name, const pr_netaddr_t *addr) {
   }
 
   port = ntohs(pr_netaddr_get_port(addr));
+  if (port == 0) {
+    port = server_port;
+  }
   namebind = pr_namebind_find(name, addr, port, FALSE);
   if (namebind == NULL) {
     errno = ENOENT;
@@ -1220,8 +1227,7 @@ static int init_inetd_bindings(void) {
      */
 
     serv->listen = main_server->listen;
-    register_cleanup(serv->listen->pool, &serv->listen, server_cleanup_cb,
-      server_cleanup_cb);
+    register_cleanup2(serv->listen->pool, &serv->listen, server_cleanup_cb);
 
     is_default = FALSE;
     default_server = get_param_ptr(serv->conf, "DefaultServer", FALSE);
@@ -1239,6 +1245,49 @@ static int init_inetd_bindings(void) {
   return 0;
 }
 
+static unsigned int process_serveralias(server_rec *s) {
+  unsigned namebind_count = 0;
+  config_rec *c;
+
+  /* If there is no ipbind already for this server, we cannot associate
+   * any ServerAlias-based namebinds to it.
+   */
+  if (pr_ipbind_get_server(s->addr, s->ServerPort) == NULL) {
+    return 0;
+  }
+
+  c = find_config(s->conf, CONF_PARAM, "ServerAlias", FALSE);
+  while (c != NULL) {
+    int res;
+
+    pr_signals_handle();
+
+    res = pr_namebind_create(s, c->argv[0], s->addr, s->ServerPort);
+    if (res == 0) {
+      namebind_count++;
+
+      res = pr_namebind_open(c->argv[0], s->addr, s->ServerPort);
+      if (res < 0) {
+        pr_trace_msg(trace_channel, 2,
+          "notice: unable to open namebind '%s': %s", (char *) c->argv[0],
+          strerror(errno));
+      }
+
+    } else {
+      if (errno != ENOENT) {
+        pr_trace_msg(trace_channel, 3,
+          "unable to create namebind for '%s' to %s#%u: %s",
+          (char *) c->argv[0], pr_netaddr_get_ipstr(s->addr), s->ServerPort,
+          strerror(errno));
+      }
+    }
+
+    c = find_config_next(c, c->next, CONF_PARAM, "ServerAlias", FALSE);
+  }
+
+  return namebind_count;
+}
+
 static int init_standalone_bindings(void) {
   int res = 0;
   server_rec *serv = NULL;
@@ -1248,7 +1297,6 @@ static int init_standalone_bindings(void) {
    * at all.
    */
   if (main_server->ServerPort) {
-
     /* If SocketBindTight is off, then pr_inet_create_conn() will
      * create and bind to a wildcard socket.  However, should it be an
      * IPv4 or an IPv6 wildcard socket?
@@ -1295,38 +1343,13 @@ static int init_standalone_bindings(void) {
   }
 
   for (serv = main_server->next; serv; serv = serv->next) {
-    config_rec *c;
-    int is_namebind = FALSE;
+    unsigned int namebind_count;
 
-    /* See if this server is a namebind, to be part of an existing ipbind. */
-    c = find_config(serv->conf, CONF_PARAM, "ServerAlias", FALSE);
-    while (c != NULL) {
-      pr_signals_handle();
-
-      res = pr_namebind_create(serv, c->argv[0], serv->addr, serv->ServerPort);
-      if (res == 0) {
-        is_namebind = TRUE;
-
-        res = pr_namebind_open(c->argv[0], serv->addr);
-        if (res < 0) {
-          pr_trace_msg(trace_channel, 2,
-            "notice: unable to open namebind '%s': %s", (char *) c->argv[0],
-            strerror(errno));
-        }
-
-      } else {
-        if (errno != ENOENT) {
-          pr_trace_msg(trace_channel, 3,
-            "unable to create namebind for '%s' to %s#%u: %s",
-            (char *) c->argv[0], pr_netaddr_get_ipstr(serv->addr),
-            serv->ServerPort, strerror(errno));
-        }
-      }
-
-      c = find_config_next(c, c->next, CONF_PARAM, "ServerAlias", FALSE);
-    }
-
-    if (is_namebind == TRUE) {
+    namebind_count = process_serveralias(serv);
+    if (namebind_count > 0) {
+      /* If we successfully added ServerAlias namebinds, move on to the next
+       * server.
+       */
       continue;
     }
 
@@ -1392,14 +1415,18 @@ static int init_standalone_bindings(void) {
       }
 
       serv->listen = main_server->listen;
-      register_cleanup(serv->listen->pool, &serv->listen, server_cleanup_cb,
-        server_cleanup_cb);
+      register_cleanup2(serv->listen->pool, &serv->listen, server_cleanup_cb);
 
       PR_CREATE_IPBIND(serv, serv->addr, serv->ServerPort);
       PR_OPEN_IPBIND(serv->addr, serv->ServerPort, NULL, is_default, FALSE,
         TRUE);
       PR_ADD_IPBINDS(serv);
     }
+
+    /* Process any ServerAlias directives AFTER the server's ipbind has been
+     * created/opened (Issue #932).
+     */
+    process_serveralias(serv);
   }
 
   /* Any "unclaimed" listening conns can be removed and closed. */
