@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2016 The ProFTPD Project team
+ * Copyright (c) 2001-2020 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -159,7 +159,8 @@ static int data_passive_open(const char *reason, off_t size) {
     if (session.xfer.xfer_type != STOR_UNIQUE) {
       if (size) {
         pr_response_send(R_150, _("Opening %s mode data connection for %s "
-          "(%" PR_LU " bytes)"), MODE_STRING, reason, (pr_off_t) size);
+          "(%" PR_LU " %s)"), MODE_STRING, reason, (pr_off_t) size,
+          size != 1 ? "bytes" : "byte");
 
       } else {
         pr_response_send(R_150, _("Opening %s mode data connection for %s"),
@@ -203,8 +204,7 @@ static int data_passive_open(const char *reason, off_t size) {
   xerrno = session.d->xerrno;
   pr_response_add_err(R_425, _("Unable to build data connection: %s"),
     strerror(xerrno));
-  destroy_pool(session.d->pool);
-  session.d = NULL;
+  pr_data_close2();
 
   errno = xerrno;
   return -1;
@@ -212,9 +212,8 @@ static int data_passive_open(const char *reason, off_t size) {
 
 static int data_active_open(const char *reason, off_t size) {
   conn_t *c;
-  int bind_port, rev;
+  int bind_port, rev, *root_revoke = NULL, xerrno;
   const pr_netaddr_t *bind_addr = NULL;
-  unsigned char *root_revoke = NULL;
 
   if (session.c->remote_addr == NULL) {
     /* An opened but unconnected connection? */
@@ -243,7 +242,14 @@ static int data_active_open(const char *reason, off_t size) {
   bind_port = session.c->local_port-1;
 
   root_revoke = get_param_ptr(TOPLEVEL_CONF, "RootRevoke", FALSE);
-  if (root_revoke != NULL) {
+  if (root_revoke == NULL) {
+    /* In the absence of any explicit RootRevoke, the default behavior is to
+     * change the source port.  This means we are technically noncompliant,
+     * but most clients do not enforce this behavior.
+     */
+    bind_port = INPORT_ANY;
+
+  } else {
     /* A RootRevoke value of 0 indicates 'false', 1 indicates 'true', and
      * 2 indicates 'NonCompliantActiveTransfer'.  We change the source port for
      * a RootRevoke value of 2, and for a value of 1, we make sure that
@@ -270,11 +276,10 @@ static int data_active_open(const char *reason, off_t size) {
 
   session.d = pr_inet_create_conn(session.pool, -1, bind_addr, bind_port, TRUE);
   if (session.d == NULL) {
-    int xerrno = errno;
+    xerrno = errno;
 
     pr_response_add_err(R_425, _("Unable to build data connection: %s"),
       strerror(xerrno));
-    session.d = NULL;
 
     errno = xerrno;
     return -1;
@@ -326,7 +331,7 @@ static int data_active_open(const char *reason, off_t size) {
 
   if (pr_inet_connect(session.d->pool, session.d, &session.data_addr,
       session.data_port) < 0) {
-    int xerrno = session.d->xerrno;
+    xerrno = session.d->xerrno;
 
     pr_log_debug(DEBUG6,
       "Error connecting to %s#%u for active data transfer: %s",
@@ -334,9 +339,7 @@ static int data_active_open(const char *reason, off_t size) {
       strerror(xerrno));
     pr_response_add_err(R_425, _("Unable to build data connection: %s"),
       strerror(xerrno));
-
-    destroy_pool(session.d->pool);
-    session.d = NULL;
+    pr_data_close2();
 
     errno = xerrno;
     return -1;
@@ -356,7 +359,8 @@ static int data_active_open(const char *reason, off_t size) {
     if (session.xfer.xfer_type != STOR_UNIQUE) {
       if (size) {
         pr_response_send(R_150, _("Opening %s mode data connection for %s "
-          "(%" PR_LU " bytes)"), MODE_STRING, reason, (pr_off_t) size);
+          "(%" PR_LU " %s)"), MODE_STRING, reason, (pr_off_t) size,
+          size != 1 ? "bytes" : "byte");
 
       } else {
         pr_response_send(R_150, _("Opening %s mode data connection for %s"),
@@ -395,10 +399,10 @@ static int data_active_open(const char *reason, off_t size) {
 
   pr_response_add_err(R_425, _("Unable to build data connection: %s"),
     strerror(session.d->xerrno));
-  errno = session.d->xerrno;
+  xerrno = session.d->xerrno;
+  pr_data_close2();
 
-  destroy_pool(session.d->pool);
-  session.d = NULL;
+  errno = xerrno;
   return -1;
 }
 
@@ -567,20 +571,26 @@ int pr_data_open(char *filename, char *reason, int direction, off_t size) {
     struct sigaction act;
 
     if (pr_netio_postopen(session.d->instrm) < 0) {
+      int xerrno;
+
       pr_response_add_err(R_425, _("Unable to build data connection: %s"),
         strerror(session.d->xerrno));
-      destroy_pool(session.d->pool);
-      errno = session.d->xerrno;
-      session.d = NULL;
+      xerrno = session.d->xerrno;
+      pr_data_close2();
+
+      errno = xerrno;
       return -1;
     }
 
     if (pr_netio_postopen(session.d->outstrm) < 0) {
+      int xerrno;
+
       pr_response_add_err(R_425, _("Unable to build data connection: %s"),
         strerror(session.d->xerrno));
-      destroy_pool(session.d->pool);
-      errno = session.d->xerrno;
-      session.d = NULL;
+      xerrno = session.d->xerrno;
+      pr_data_close2();
+
+      errno = xerrno;
       return -1;
     }
 
@@ -644,11 +654,10 @@ int pr_data_open(char *filename, char *reason, int direction, off_t size) {
   return res;
 }
 
-/* close == successful transfer */
-void pr_data_close(int quiet) {
+void pr_data_close2(void) {
   nstrm = NULL;
 
-  if (session.d) {
+  if (session.d != NULL) {
     pr_inet_lingering_close(session.pool, session.d, timeout_linger);
     session.d = NULL;
   }
@@ -667,8 +676,13 @@ void pr_data_close(int quiet) {
   session.sf_flags &= (SF_ALL^SF_PASSIVE);
   session.sf_flags &= (SF_ALL^(SF_ABORT|SF_XFER|SF_PASSIVE|SF_ASCII_OVERRIDE));
   pr_session_set_idle();
+}
 
-  if (!quiet) {
+/* close == successful transfer */
+void pr_data_close(int quiet) {
+  pr_data_close2();
+
+  if (quiet == FALSE) {
     pr_response_add(R_226, _("Transfer complete"));
   }
 }
@@ -684,7 +698,7 @@ void pr_data_close(int quiet) {
  */
 void pr_data_cleanup(void) {
   /* sanity check */
-  if (session.d) {
+  if (session.d != NULL) {
     pr_inet_lingering_close(session.pool, session.d, timeout_linger);
     session.d = NULL;
   }
@@ -711,7 +725,7 @@ void pr_data_abort(int err, int quiet) {
     strerror(err), err, quiet ? "true" : "false",
     true_abort ? "true" : "false");
 
-  if (session.d) {
+  if (session.d != NULL) {
     if (true_abort == FALSE) {
       pr_inet_lingering_close(session.pool, session.d, timeout_linger);
 
@@ -879,7 +893,7 @@ void pr_data_abort(int err, int quiet) {
     if (msg == NULL &&
         (msg = strerror(err)) == NULL ) {
 
-      if (snprintf(msgbuf, sizeof(msgbuf),
+      if (pr_snprintf(msgbuf, sizeof(msgbuf),
           _("Unknown or out of range errno [%d]"), err) > 0)
 	msg = msgbuf;
     }
@@ -893,6 +907,11 @@ void pr_data_abort(int err, int quiet) {
     if (true_abort == FALSE) {
       pr_response_add_err(respcode, _("Transfer aborted. %s"), msg ? msg : "");
     }
+
+    /* Forcibly clear the data-transfer instigating command pool from the
+     * Response API.
+     */
+    pr_response_set_pool(session.pool);
   }
 
   if (true_abort) {
@@ -925,6 +944,7 @@ static void poll_ctrl(void) {
     res = pr_cmd_read(&cmd);
     if (res < 0) {
       int xerrno;
+
 #if defined(ECONNABORTED)
       xerrno = ECONNABORTED;
 #elif defined(ENOTCONN)
@@ -993,8 +1013,8 @@ static void poll_ctrl(void) {
 
         pr_response_flush(&resp_err_list);
 
-        destroy_pool(cmd->pool);
         pr_response_set_pool(resp_pool);
+        destroy_pool(cmd->pool);
 
       /* We don't want to actually dispatch the NOOP command, since that
        * would overwrite the scoreboard with the NOOP state; admins probably
@@ -1019,13 +1039,14 @@ static void poll_ctrl(void) {
 
         pr_response_flush(&resp_list);
 
-        destroy_pool(cmd->pool);
         pr_response_set_pool(resp_pool);
+        destroy_pool(cmd->pool);
 
       } else {
         char *title_buf = NULL;
-        int title_len = -1;
-        const char *sce_cmd = NULL, *sce_cmd_arg = NULL;
+        int curr_cmd_id = 0, title_len = -1;
+        const char *curr_cmd = NULL, *sce_cmd = NULL, *sce_cmd_arg = NULL;
+        cmd_rec *curr_cmd_rec = NULL;
 
         pr_trace_msg(trace_channel, 5,
           "client sent '%s' command during data transfer, dispatching",
@@ -1037,6 +1058,9 @@ static void poll_ctrl(void) {
           pr_proctitle_get(title_buf, title_len + 1); 
         }
 
+        curr_cmd = session.curr_cmd;
+        curr_cmd_id = session.curr_cmd_id;
+        curr_cmd_rec = session.curr_cmd_rec;
         sce_cmd = pr_scoreboard_entry_get(PR_SCORE_CMD);
         sce_cmd_arg = pr_scoreboard_entry_get(PR_SCORE_CMD_ARG);
 
@@ -1052,6 +1076,9 @@ static void poll_ctrl(void) {
         }
 
         destroy_pool(cmd->pool);
+        session.curr_cmd = curr_cmd;
+        session.curr_cmd_id = curr_cmd_id;
+        session.curr_cmd_rec = curr_cmd_rec;
       }
 
     } else {
@@ -1135,7 +1162,7 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
         while (len < 0) {
           int xerrno = errno;
  
-          if (xerrno == EAGAIN) {
+          if (xerrno == EAGAIN || xerrno == EINTR) {
             /* Since our socket is in non-blocking mode, read(2) can return
              * EAGAIN if there is no data yet for us.  Handle this by
              * delaying temporarily, then trying again.
@@ -1170,6 +1197,9 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
         }
 
         if (len > 0) {
+          pr_trace_msg(trace_channel, 19, "read %d %s from network", len,
+            len != 1 ? "bytes" : "byte");
+
           buflen += len;
 
           if (timeout_stalled) {
@@ -1197,12 +1227,10 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
               buflen > 1) {
             size_t outlen = 0;
 
-            /* Use a temporary pool for the CRLF conversion, lest the
-             * session.xfer.p pool grow quite large while downloading a large
-             * file for ASCII conversion (Bug#4277).
-             */
-            tmp_pool = make_sub_pool(session.xfer.p);
-            pr_pool_tag(tmp_pool, "ASCII download");
+            if (tmp_pool == NULL) {
+              tmp_pool = make_sub_pool(session.xfer.p);
+              pr_pool_tag(tmp_pool, "ASCII upload");
+            }
 
             res = pr_ascii_ftp_from_crlf(tmp_pool, buf, buflen, &buf, &outlen);
             if (res < 0) {
@@ -1257,7 +1285,7 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
       while (len < 0) {
         int xerrno = errno;
 
-        if (xerrno == EAGAIN) {
+        if (xerrno == EAGAIN || xerrno == EINTR) {
           /* Since our socket is in non-blocking mode, read(2) can return
            * EAGAIN if there is no data yet for us.  Handle this by
            * delaying temporarily, then trying again.
@@ -1273,6 +1301,9 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
       }
 
       if (len > 0) {
+        pr_trace_msg(trace_channel, 19, "read %d %s from network", len,
+          len != 1 ? "bytes" : "byte");
+
         if (data_first_byte_read == FALSE) {
           if (pr_trace_get_level(timing_channel)) {
             unsigned long elapsed_ms;
@@ -1303,6 +1334,7 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
       int bwrote = 0;
       int buflen = cl_size;
       unsigned int xferbuflen;
+      char *xfer_buf = NULL;
 
       pr_signals_handle();
 
@@ -1326,12 +1358,10 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
         char *out = NULL;
         size_t outlen = 0;
 
-        /* Use a temporary pool for the CRLF conversion, lest the
-         * session.xfer.p pool grow quite large while downloading a large
-         * file for ASCII conversion (Bug#4277).
-         */
-        tmp_pool = make_sub_pool(session.xfer.p);
-        pr_pool_tag(tmp_pool, "ASCII upload");
+        if (tmp_pool == NULL) {
+          tmp_pool = make_sub_pool(session.xfer.p);
+          pr_pool_tag(tmp_pool, "ASCII download");
+        }
 
         /* Scan the internal buffer, looking for LFs with no preceding CRs.
          * Add CRs (and expand the internal buffer) as necessary. xferbuflen
@@ -1345,6 +1375,7 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
             strerror(errno));
 
         } else {
+          xfer_buf = session.xfer.buf;
           session.xfer.buf = out;
           session.xfer.buflen = xferbuflen = outlen;
         }
@@ -1354,7 +1385,7 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
       while (bwrote < 0) {
         int xerrno = errno;
 
-        if (xerrno == EAGAIN) {
+        if (xerrno == EAGAIN || xerrno == EINTR) {
           /* Since our socket is in non-blocking mode, write(2) can return
            * EAGAIN if there is not enough from for our data yet.  Handle
            * this by delaying temporarily, then trying again.
@@ -1368,11 +1399,20 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
         }
 
         destroy_pool(tmp_pool);
+        if (xfer_buf != NULL) {
+          /* Free up the malloc'd memory. */
+          free(session.xfer.buf);
+          session.xfer.buf = xfer_buf;
+        }
+
         errno = xerrno;
         return -1;
       }
 
       if (bwrote > 0) {
+        pr_trace_msg(trace_channel, 19, "wrote %d %s to network", bwrote,
+          bwrote != 1 ? "bytes" : "byte");
+
         if (data_first_byte_written == FALSE) {
           if (pr_trace_get_level(timing_channel)) {
             unsigned long elapsed_ms;
@@ -1396,6 +1436,14 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
         cl_buf += buflen;
         total += buflen;
       }
+
+      if (xfer_buf != NULL) {
+        /* Yes, we are using malloc et al here, rather than the memory pools.
+         * See Bug#4352 for details.
+         */
+        free(session.xfer.buf);
+        session.xfer.buf = xfer_buf;
+      }
     }
 
     len = total;
@@ -1419,7 +1467,7 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
   return (len < 0 ? -1 : len);
 }
 
-#ifdef HAVE_SENDFILE
+#if defined(HAVE_SENDFILE)
 /* pr_data_sendfile() actually transfers the data on the data connection.
  * ASCII translation is not performed.
  * return 0 if reading and data connection closes, or -1 if error
@@ -1427,10 +1475,10 @@ int pr_data_xfer(char *cl_buf, size_t cl_size) {
 pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
   int flags, error;
   pr_sendfile_t len = 0, total = 0;
-#if defined(HAVE_AIX_SENDFILE)
+# if defined(HAVE_AIX_SENDFILE)
   struct sf_parms parms;
   int rc;
-#endif /* HAVE_AIX_SENDFILE */
+# endif /* HAVE_AIX_SENDFILE */
 
   if (offset == NULL ||
       count == 0) {
@@ -1461,7 +1509,7 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
   }
 
   for (;;) {
-#if defined(HAVE_LINUX_SENDFILE) || defined(HAVE_SOLARIS_SENDFILE)
+# if defined(HAVE_LINUX_SENDFILE) || defined(HAVE_SOLARIS_SENDFILE)
     off_t orig_offset = *offset;
 
     /* Linux semantics are fairly straightforward in a glibc 2.x world:
@@ -1479,27 +1527,42 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
      * times.  How annoying.
      */
 
-#if defined(HAVE_LINUX_SENDFILE)
+#  if defined(HAVE_LINUX_SENDFILE)
     if (count > INT_MAX) {
       count = INT_MAX;
     }
-#elif defined(HAVE_SOLARIS_SENDFILE)
-# if SIZEOF_SIZE_T == SIZEOF_INT
+#  elif defined(HAVE_SOLARIS_SENDFILE)
+#   if SIZEOF_SIZE_T == SIZEOF_INT
     if (count > INT_MAX) {
       count = INT_MAX;
     }
-# elif SIZEOF_SIZE_T == SIZEOF_LONG
+#   elif SIZEOF_SIZE_T == SIZEOF_LONG
     if (count > LONG_MAX) {
       count = LONG_MAX;
     }
-# elif SIZEOF_SIZE_T == SIZEOF_LONG_LONG
+#   elif SIZEOF_SIZE_T == SIZEOF_LONG_LONG
     if (count > LLONG_MAX) {
       count = LLONG_MAX;
     }
-# endif
-#endif /* !HAVE_SOLARIS_SENDFILE */
+#   endif
+#  endif /* !HAVE_SOLARIS_SENDFILE */
 
+    errno = 0;
     len = sendfile(PR_NETIO_FD(session.d->outstrm), retr_fd, offset, count);
+
+    /* If no data could be written (e.g. the file was truncated), we're
+     * done (Bug#4318).
+     */
+    if (len == 0) {
+      if (errno != EINTR &&
+          errno != EAGAIN) {
+        break;
+      }
+
+      /* Handle our interrupting signal, and continue. */
+      pr_signals_handle();
+    }
+
     if (len != -1 &&
         len < count) {
       /* Under Linux semantics, this occurs when a signal has interrupted
@@ -1545,7 +1608,7 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
       len = *offset - orig_offset;
       *offset = orig_offset;
 
-#elif defined(HAVE_BSD_SENDFILE)
+# elif defined(HAVE_BSD_SENDFILE)
     /* BSD semantics for sendfile are flexible...it'd be nice if we could
      * standardize on something like it.  The semantics are:
      *
@@ -1562,24 +1625,24 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
      *  value.
      */
 
-#if SIZEOF_SIZE_T == SIZEOF_INT
+#  if SIZEOF_SIZE_T == SIZEOF_INT
     if (count > UINT_MAX) {
       count = UINT_MAX;
     }
-#elif SIZEOF_SIZE_T == SIZEOF_LONG
+#  elif SIZEOF_SIZE_T == SIZEOF_LONG
     if (count > ULONG_MAX) {
       count = ULONG_MAX;
     }
-#elif SIZEOF_SIZE_T == SIZEOF_LONG_LONG
+#  elif SIZEOF_SIZE_T == SIZEOF_LONG_LONG
     if (count > ULLONG_MAX) {
       count = ULLONG_MAX;
     }
-#endif
+#  endif
 
     if (sendfile(retr_fd, PR_NETIO_FD(session.d->outstrm), *offset, count,
         NULL, &len, 0) == -1) {
 
-#elif defined(HAVE_MACOSX_SENDFILE)
+# elif defined(HAVE_MACOSX_SENDFILE)
     off_t orig_len = count;
     int res;
 
@@ -1593,7 +1656,7 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
     len = orig_len;
 
     if (res < 0) {
-#elif defined(HAVE_AIX_SENDFILE)
+# elif defined(HAVE_AIX_SENDFILE)
 
     memset(&parms, 0, sizeof(parms));
 
@@ -1605,8 +1668,9 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
     len = (int) parms.bytes_sent;
 
     if (rc < -1 || rc == 1) {
-
-#endif /* HAVE_AIX_SENDFILE */
+# else
+    if (FALSE) {
+# endif /* HAVE_AIX_SENDFILE */
 
       /* IMO, BSD's semantics are warped.  Apparently, since we have our
        * alarms tagged SA_INTERRUPT (allowing system calls to be
