@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2016 The ProFTPD Project team
+ * Copyright (c) 2001-2020 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -479,8 +479,7 @@ static config_rec *recur_match_path(pool *p, xaset_t *s, char *path) {
           !(tmp_path[path_len-2] == '/' && tmp_path[path_len-1] == '*')) {
 
         /* Trim a trailing path separator, if present. */
-        if (path_len > 1 &&
-            *tmp_path && 
+        if (*tmp_path &&
             *(tmp_path + path_len - 1) == '/') {
           *(tmp_path + path_len - 1) = '\0';
           path_len--;
@@ -1133,12 +1132,11 @@ static int check_ip_access(xaset_t *set, char *name) {
 static int check_limit_allow(config_rec *c, cmd_rec *cmd) {
   unsigned char *allow_all = NULL;
 
-  /* If session.groups is null, this means no authentication
-   * attempt has been made, so we simply check for the
-   * very existance of an AllowGroup, and assume (for now) it's
-   * allowed.  This works because later calls to check_limit_allow()
-   * WILL have filled in the group members and we can truely check
-   * group membership at that time.  Same goes for AllowUser.
+  /* If session.groups is null, this means no authentication attempt has been
+   * made, so we simply check for the very existence of an AllowGroup, and
+   * assume (for now) it's allowed.  This works because later calls to
+   * check_limit_allow() WILL have filled in the group members and we can
+   * truly check group membership at that time.  Same goes for AllowUser.
    */
 
   if (!session.user) {
@@ -2534,7 +2532,7 @@ int fixup_servers(xaset_t *list) {
               /* Create the bind record using the IPv4-mapped IPv6 version of
                * this address.
                */
-              snprintf(ipbuf, INET6_ADDRSTRLEN, "::ffff:%s", ipstr);
+              pr_snprintf(ipbuf, INET6_ADDRSTRLEN, "::ffff:%s", ipstr);
               ipstr = pstrdup(s->pool, ipbuf);
             }
           }
@@ -2636,7 +2634,7 @@ int fixup_servers(xaset_t *list) {
 }
 
 static void set_tcp_bufsz(server_rec *s) {
-  int sockfd;
+  int proto = -1, sockfd;
   socklen_t optlen = 0;
   struct protoent *p = NULL;
 
@@ -2645,6 +2643,14 @@ static void set_tcp_bufsz(server_rec *s) {
 #endif
 
   p = getprotobyname("tcp");
+  if (p != NULL) {
+    proto = p->p_proto;
+  }
+
+#ifdef HAVE_ENDPROTOENT
+  endprotoent();
+#endif
+
   if (p == NULL) {
 #ifndef PR_TUNABLE_RCVBUFSZ
     s->tcp_rcvbuf_len = tcp_rcvbufsz = PR_TUNABLE_DEFAULT_RCVBUFSZ;
@@ -2662,18 +2668,18 @@ static void set_tcp_bufsz(server_rec *s) {
     pr_log_debug(DEBUG4, "using default TCP receive/send buffer sizes");
 
 #ifndef PR_TUNABLE_XFER_BUFFER_SIZE
-  /* Choose the smaller of the two TCP buffer sizes as the overall transfer
-   * size (for use by the data transfer layer).
-   */
-   xfer_bufsz = tcp_sndbufsz < tcp_rcvbufsz ? tcp_sndbufsz : tcp_rcvbufsz;
+    /* Choose the smaller of the two TCP buffer sizes as the overall transfer
+     * size (for use by the data transfer layer).
+     */
+     xfer_bufsz = tcp_sndbufsz < tcp_rcvbufsz ? tcp_sndbufsz : tcp_rcvbufsz;
 #else
-  xfer_bufsz = PR_TUNABLE_XFER_BUFFER_SIZE;
+    xfer_bufsz = PR_TUNABLE_XFER_BUFFER_SIZE;
 #endif /* PR_TUNABLE_XFER_BUFFER_SIZE */
 
     return;
   }
 
-  sockfd = socket(AF_INET, SOCK_STREAM, p->p_proto);
+  sockfd = socket(AF_INET, SOCK_STREAM, proto);
   if (sockfd < 0) {
     s->tcp_rcvbuf_len = tcp_rcvbufsz = PR_TUNABLE_DEFAULT_RCVBUFSZ;
     s->tcp_sndbuf_len = tcp_sndbufsz = PR_TUNABLE_DEFAULT_SNDBUFSZ;
@@ -2848,7 +2854,7 @@ char *get_context_name(cmd_rec *cmd) {
        * handle their own arbitrary configuration contexts.
        */
       memset(cbuf, '\0', sizeof(cbuf));
-      snprintf(cbuf, sizeof(cbuf), "%d", cmd->config->config_type);
+      pr_snprintf(cbuf, sizeof(cbuf), "%d", cmd->config->config_type);
       return cbuf;
   }
 }
