@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001-2013 The ProFTPD Project team
+ * Copyright (c) 2001-2020 The ProFTPD Project team
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,9 +22,7 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/* ProFTPD Controls command-line client
- * $Id: ftpdctl.c,v 1.20 2013-07-16 21:36:49 castaglia Exp $
- */
+/* ProFTPD Controls command-line client */
 
 #include "conf.h"
 #include "privs.h"
@@ -37,6 +35,7 @@ static const char *program = "ftpdctl";
 
 /* NOTE: these empty stubs are needed for proper linking.  What a mess. */
 
+unsigned char is_master;
 session_t session;
 server_rec *main_server = NULL;
 
@@ -67,10 +66,30 @@ int pr_event_listening(const char *event) {
   return -1;
 }
 
+void pr_fs_fadvise(int fd, off_t off, off_t len, int advice) {
+}
+
 int pr_fs_get_usable_fd(int fd) {
   return -1;
 }
 
+struct tm *pr_localtime(pool *p, const time_t *t) {
+  struct tm *tm;
+
+#if defined(HAVE_LOCALTIME_R)
+  if (p == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  tm = localtime_r(t, pcalloc(p, sizeof(struct tm)));
+#else
+  tm = localtime(t);
+#endif /* HAVE_LOCALTIME_R */
+
+  return tm;
+}
+
 int pr_privs_root(const char *file, int lineno) {
   return 0;
 }
@@ -79,6 +98,10 @@ int pr_privs_relinquish(const char *file, int lineno) {
   return 0;
 }
 
+long pr_random_next(long min, long max) {
+  return 1;
+}
+
 void pr_signals_block(void) {
 }
 
@@ -88,12 +111,23 @@ void pr_signals_unblock(void) {
 void pr_signals_handle(void) {
 }
 
+int pr_snprintf(char *buf, size_t bufsz, const char *fmt, ...) {
+  va_list msg;
+  int res;
+
+  va_start(msg, fmt);
+  res = pr_vsnprintf(buf, bufsz, fmt, msg);
+  va_end(msg);
+
+  return res;
+}
+
 pr_table_t *pr_table_alloc(pool *p, int flags) {
   errno = EPERM;
   return NULL;
 }
 
-int pr_table_add(pr_table_t *tab, const char *k, void *v, size_t sz) {
+int pr_table_add(pr_table_t *tab, const char *k, const void *v, size_t sz) {
   errno = EPERM;
   return -1;
 }
@@ -118,18 +152,19 @@ int pr_table_free(pr_table_t *tab) {
   return -1;
 }
 
-void *pr_table_get(pr_table_t *tab, const char *k, size_t *sz) {
+const void *pr_table_get(pr_table_t *tab, const char *k, size_t *sz) {
   errno = EPERM;
   return NULL;
 }
 
-int pr_table_set(pr_table_t *tab, const char *k, void *v, size_t sz) {
+const void *pr_table_remove(pr_table_t *tab, const char *k, size_t *sz) {
   errno = EPERM;
-  return -1;
+  return NULL;
 }
 
-struct tm *pr_localtime(pool *p, const time_t *t) {
-  return localtime(t);
+int pr_table_set(pr_table_t *tab, const char *k, const void *v, size_t sz) {
+  errno = EPERM;
+  return -1;
 }
 
 int pr_trace_msg(const char *channel, int level, const char *fmt, ...) {
@@ -137,6 +172,10 @@ int pr_trace_msg(const char *channel, int level, const char *fmt, ...) {
   return -1;
 }
 
+int pr_vsnprintf(char *buf, size_t bufsz, const char *fmt, va_list msg) {
+  return vsnprintf(buf, bufsz, fmt, msg);
+}
+
 int sstrncpy(char *dst, const char *src, size_t n) {
   register char *d = dst;
   int res = 0;
@@ -169,7 +208,7 @@ static RETSIGTYPE sig_pipe(int sig) {
 }
 
 static void usage(void) {
-  fprintf(stdout, "usage: %s [options]\n", program);
+  fprintf(stdout, "usage: %s [options] action [...]\n", program);
   fprintf(stdout, "  -h\tdisplays this message\n");
   fprintf(stdout, "  -s\tspecify an alternate local socket\n");
   fprintf(stdout, "  -v\tdisplays more verbose information\n");
@@ -189,12 +228,6 @@ int main(int argc, char *argv[]) {
   pool *ctl_pool = NULL;
   array_header *reqargv = NULL;
 
-  /* Make sure we were called with at least one argument. */
-  if (argc-1 < 1) {
-    fprintf(stdout, "%s: missing required arguments\n", program);
-    exit(1);
-  }
-
   /* Set the POSIXLY_CORRECT environment variable, so that control handlers
    * can themselves have optional flags.
    */
@@ -231,6 +264,12 @@ int main(int argc, char *argv[]) {
     }
   }
 
+  /* Make sure we were called with at least one argument. */
+  if (argv[optind] == NULL) {
+    fprintf(stdout, "%s: missing required action\n", program);
+    exit(1);
+  }
+
   signal(SIGPIPE, sig_pipe);
 
   /* Allocate some memory for proftpd objects. */
