@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2013 The ProFTPD Project team
+ * Copyright (c) 2001-2020 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,25 +24,33 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/* ProFTPD logging support.
- * $Id: log.c,v 1.121 2013/10/13 23:46:43 castaglia Exp $
- */
+/* ProFTPD logging support. */
 
 #include "conf.h"
 
-/* Max path length plus 64 bytes for additional info. */
-#define LOGBUFFER_SIZE		(PR_TUNABLE_PATH_MAX + 64)
+#ifdef HAVE_EXECINFO_H
+# include <execinfo.h>
+#endif
+
+#define LOGBUFFER_SIZE		(PR_TUNABLE_PATH_MAX * 4)
+
+/* From src/main.c */
+extern unsigned char is_master;
 
 static int syslog_open = FALSE;
 static int syslog_discard = FALSE;
+static unsigned long log_opts = PR_LOG_OPT_DEFAULT;
 static int logstderr = TRUE;
-static int debug_level = DEBUG0;	/* Default is no debug logging */
+static int debug_level = DEBUG0;
+static int default_level = PR_LOG_NOTICE;
 static int facility = LOG_DAEMON;
 static int set_facility = -1;
 static char systemlog_fn[PR_TUNABLE_PATH_MAX] = {'\0'};
 static char systemlog_host[256] = {'\0'};
 static int systemlog_fd = -1;
 
+static const char *trace_channel = "log";
+
 int syslog_sockfd = -1;
 
 #ifdef PR_USE_NONBLOCKING_LOG_OPEN
@@ -59,12 +67,13 @@ static int fd_set_block(int fd) {
 int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
   int res;
   pool *tmp_pool = NULL;
-  char *tmp = NULL, *lf;
+  char *ptr = NULL, *lf;
   unsigned char have_stat = FALSE, *allow_log_symlinks = NULL;
   struct stat st;
 
   /* Sanity check */
-  if (!log_file || !log_fd) {
+  if (log_file == NULL ||
+      log_fd == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -74,8 +83,8 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
   pr_pool_tag(tmp_pool, "log_openfile() tmp pool");
   lf = pstrdup(tmp_pool, log_file);
 
-  tmp = strrchr(lf, '/');
-  if (tmp == NULL) {
+  ptr = strrchr(lf, '/');
+  if (ptr == NULL) {
     pr_log_debug(DEBUG0, "inappropriate log file: %s", lf);
     destroy_pool(tmp_pool);
 
@@ -86,7 +95,9 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
   /* Set the path separator to zero, in order to obtain the directory
    * name, so that checks of the directory may be made.
    */
-  *tmp = '\0';
+  if (ptr != lf) {
+    *ptr = '\0';
+  }
 
   if (stat(lf, &st) < 0) {
     int xerrno = errno;
@@ -117,7 +128,9 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
   /* Restore the path separator so that checks on the file itself may be
    * done.
    */
-  *tmp = '/';
+  if (ptr != lf) {
+    *ptr = '/';
+  }
 
   allow_log_symlinks = get_param_ptr(main_server->conf, "AllowLogSymlinks",
     FALSE);
@@ -214,8 +227,9 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
 
     /* Stat the file using the descriptor, not the path */
     if (!have_stat &&
-        fstat(*log_fd, &st) != -1)
+        fstat(*log_fd, &st) != -1) {
       have_stat = TRUE;
+    }
 
     if (!have_stat ||
         S_ISLNK(st.st_mode)) {
@@ -241,11 +255,44 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
 
     *log_fd = open(lf, flags, log_mode);
     if (*log_fd < 0) {
+      int xerrno = errno;
+
       destroy_pool(tmp_pool);
+      errno = xerrno;
       return -1;
     }
   }
 
+  /* Make sure we're dealing with an expected file type (i.e. NOT a
+   * directory).
+   */
+  if (fstat(*log_fd, &st) < 0) {
+    int xerrno = errno;
+
+    pr_log_debug(DEBUG0, "error: unable to stat %s (fd %d): %s", lf, *log_fd,
+      strerror(xerrno));
+
+    close(*log_fd);
+    *log_fd = -1;
+    destroy_pool(tmp_pool);
+
+    errno = xerrno;
+    return -1;
+  }
+
+  if (S_ISDIR(st.st_mode)) {
+    int xerrno = EISDIR;
+
+    pr_log_debug(DEBUG0, "error: unable to use %s: %s", lf, strerror(xerrno));
+
+    close(*log_fd);
+    *log_fd = -1;
+    destroy_pool(tmp_pool);
+
+    errno = xerrno;
+    return -1;
+  }
+
   /* Find a usable fd for the just-opened log fd. */
   if (*log_fd <= STDERR_FILENO) {
     res = pr_fs_get_usable_fd(*log_fd);
@@ -264,9 +311,14 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
       *log_fd, strerror(errno));
   }
 
+  /* Advise the platform that we will be treating this log file as
+   * write-only data.
+   */
+  pr_fs_fadvise(*log_fd, 0, 0, PR_FS_FADVISE_DONTNEED);
+
 #ifdef PR_USE_NONBLOCKING_LOG_OPEN
   /* Return the fd to blocking mode. */
-  fd_set_block(*log_fd);
+  (void) fd_set_block(*log_fd);
 #endif /* PR_USE_NONBLOCKING_LOG_OPEN */
 
   destroy_pool(tmp_pool);
@@ -275,7 +327,8 @@ int pr_log_openfile(const char *log_file, int *log_fd, mode_t log_mode) {
 
 int pr_log_vwritefile(int logfd, const char *ident, const char *fmt,
     va_list msg) {
-  char buf[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
+  pool *tmp_pool;
+  char buf[LOGBUFFER_SIZE] = {'\0'};
   struct timeval now;
   struct tm *tm = NULL;
   size_t buflen, len;
@@ -286,9 +339,16 @@ int pr_log_vwritefile(int logfd, const char *ident, const char *fmt,
     return -1;
   }
 
+  tmp_pool = make_sub_pool(permanent_pool);
+  pr_pool_tag(tmp_pool, "Log message pool");
+
   gettimeofday(&now, NULL);
-  tm = pr_localtime(NULL, (const time_t *) &(now.tv_sec));
+  tm = pr_localtime(tmp_pool, (const time_t *) &(now.tv_sec));
   if (tm == NULL) {
+    int xerrno = errno;
+
+    destroy_pool(tmp_pool);
+    errno = xerrno;
     return -1;
   }
 
@@ -300,17 +360,17 @@ int pr_log_vwritefile(int logfd, const char *ident, const char *fmt,
   /* Convert microsecs to millisecs. */
   millis = now.tv_usec / 1000;
 
-  len = snprintf(buf + buflen, sizeof(buf) - len, ",%03lu ", millis);
+  len = pr_snprintf(buf + buflen, sizeof(buf) - len, ",%03lu ", millis);
   buflen += len;
 
   /* Prepend a small header */
-  len = snprintf(buf + buflen, sizeof(buf) - buflen, "%s[%u]: ", ident,
+  len = pr_snprintf(buf + buflen, sizeof(buf) - buflen, "%s[%u]: ", ident,
     (unsigned int) (session.pid ? session.pid : getpid()));
   buflen += len;
   buf[sizeof(buf)-1] = '\0';
 
   /* Affix the message */
-  len = vsnprintf(buf + buflen, sizeof(buf) - buflen - 1, fmt, msg);
+  len = pr_vsnprintf(buf + buflen, sizeof(buf) - buflen - 1, fmt, msg);
   buflen += len;
   buf[sizeof(buf)-1] = '\0';
 
@@ -318,11 +378,15 @@ int pr_log_vwritefile(int logfd, const char *ident, const char *fmt,
     buf[buflen++] = '\n';
 
   } else {
+    buf[sizeof(buf)-5] = '.';
+    buf[sizeof(buf)-4] = '.';
+    buf[sizeof(buf)-3] = '.';
     buf[sizeof(buf)-2] = '\n';
-    buflen++;
+    buflen = sizeof(buf)-1;
   }
 
   pr_log_event_generate(PR_LOG_TYPE_UNSPEC, logfd, -1, buf, buflen);
+  destroy_pool(tmp_pool);
 
   while (write(logfd, buf, buflen) < 0) {
     if (errno == EINTR) {
@@ -369,8 +433,14 @@ int log_opensyslog(const char *fn) {
     pr_closelog(syslog_sockfd);
 
     syslog_sockfd = pr_openlog("proftpd", LOG_NDELAY|LOG_PID, facility);
-    if (syslog_sockfd < 0)
+    if (syslog_sockfd < 0) {
+      int xerrno = errno;
+
+      (void) pr_trace_msg(trace_channel, 1,
+        "error opening syslog fd: %s", strerror(xerrno));
+      errno = xerrno;
       return -1;
+    }
 
     /* Find a usable fd for the just-opened socket fd. */
     if (syslog_sockfd <= STDERR_FILENO) {
@@ -381,7 +451,7 @@ int log_opensyslog(const char *fn) {
       }
     }
 
-    fcntl(syslog_sockfd, F_SETFD, FD_CLOEXEC);
+    (void) fcntl(syslog_sockfd, F_SETFD, FD_CLOEXEC);
     systemlog_fd = -1;
 
   } else if ((res = pr_log_openfile(systemlog_fn, &systemlog_fd,
@@ -417,28 +487,35 @@ void log_discard(void) {
 }
 
 static void log_write(int priority, int f, char *s, int discard) {
-  unsigned int *max_priority = NULL;
+  int max_priority = 0, *ptr = NULL;
   char serverinfo[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
 
   memset(serverinfo, '\0', sizeof(serverinfo));
 
-  if (main_server &&
+  if (main_server != NULL &&
       main_server->ServerFQDN) {
-    pr_netaddr_t *remote_addr = pr_netaddr_get_sess_remote_addr();
-    const char *remote_name = pr_netaddr_get_sess_remote_name();
+    const pr_netaddr_t *remote_addr;
+    const char *remote_name;
+    size_t buflen = 0;
+
+    remote_addr = pr_netaddr_get_sess_remote_addr();
+    remote_name = pr_netaddr_get_sess_remote_name();
+
+    if (log_opts & PR_LOG_OPT_USE_VHOST) {
+      size_t len;
+
+      len = pr_snprintf(serverinfo, sizeof(serverinfo)-1, "%s",
+        main_server->ServerFQDN);
+      buflen += len;
+    }
 
-    snprintf(serverinfo, sizeof(serverinfo)-1, "%s", main_server->ServerFQDN);
     serverinfo[sizeof(serverinfo)-1] = '\0';
 
-    if (remote_addr && remote_name) {
-      size_t serverinfo_len;
-
-      serverinfo_len = strlen(serverinfo);
-
-      snprintf(serverinfo + serverinfo_len,
-        sizeof(serverinfo) - serverinfo_len, " (%s[%s])",
+    if (remote_addr != NULL &&
+        remote_name != NULL) {
+      pr_snprintf(serverinfo + buflen, sizeof(serverinfo) - buflen,
+        "%s(%s[%s])", (log_opts & PR_LOG_OPT_USE_VHOST) ? " " : "",
         remote_name, pr_netaddr_get_ipstr(remote_addr));
-
       serverinfo[sizeof(serverinfo)-1] = '\0';
     }
   }
@@ -446,37 +523,65 @@ static void log_write(int priority, int f, char *s, int discard) {
   if (!discard &&
       (logstderr || !main_server)) {
     char buf[LOGBUFFER_SIZE] = {'\0'};
-    size_t buflen, len;
-    struct timeval now;
-    struct tm *tm = NULL;
-    unsigned long millis;
+    size_t buflen = 0, len = 0;
+    pid_t log_pid;
+    const char *process_label = "proftpd";
 
-    gettimeofday(&now, NULL);
-    tm = pr_localtime(NULL, (const time_t *) &(now.tv_sec));
-    if (tm == NULL) {
-      return;
+    if (log_opts & PR_LOG_OPT_USE_TIMESTAMP) {
+      pool *tmp_pool;
+      struct timeval now;
+      struct tm *tm = NULL;
+      unsigned long millis;
+
+      tmp_pool = make_sub_pool(permanent_pool);
+      pr_pool_tag(tmp_pool, "Log message pool");
+
+      gettimeofday(&now, NULL);
+      tm = pr_localtime(tmp_pool, (const time_t *) &(now.tv_sec));
+      if (tm == NULL) {
+        destroy_pool(tmp_pool);
+        return;
+      }
+
+      len = strftime(buf, sizeof(buf)-1, "%Y-%m-%d %H:%M:%S", tm);
+      buflen = len;
+      buf[sizeof(buf)-1] = '\0';
+      destroy_pool(tmp_pool);
+
+      /* Convert microsecs to millisecs. */
+      millis = now.tv_usec / 1000;
+
+      len = pr_snprintf(buf + buflen, sizeof(buf) - len, ",%03lu ", millis);
+      buflen += len;
     }
 
-    len = strftime(buf, sizeof(buf)-1, "%Y-%m-%d %H:%M:%S", tm);
-    buflen = len;
     buf[sizeof(buf)-1] = '\0';
 
-    /* Convert microsecs to millisecs. */
-    millis = now.tv_usec / 1000;
+    if (log_opts & PR_LOG_OPT_USE_HOSTNAME) {
+      len = pr_snprintf(buf + buflen, sizeof(buf) - buflen, "%s ",
+        systemlog_host);
+      buflen += len;
+    }
 
-    len = snprintf(buf + buflen, sizeof(buf) - len, ",%03lu ", millis);
-    buflen += len;
-    buf[sizeof(buf)-1] = '\0';
+    log_pid = session.pid ? session.pid : getpid();
+
+    if (log_opts & PR_LOG_OPT_USE_ROLE_BASED_PROCESS_LABELS) {
+      if (is_master) {
+        process_label = "daemon";
+
+      } else {
+        process_label = "session";
+      }
+    }
 
     if (*serverinfo) {
-      len = snprintf(buf + buflen, sizeof(buf) - buflen,
-        "%s proftpd[%u] %s: %s\n", systemlog_host,
-        (unsigned int) (session.pid ? session.pid : getpid()), serverinfo, s);
+      len = pr_snprintf(buf + buflen, sizeof(buf) - buflen,
+        "%s[%u] %s: %s\n", process_label, (unsigned int) log_pid, serverinfo,
+        s);
 
     } else {
-      len = snprintf(buf + buflen, sizeof(buf) - buflen,
-        "%s proftpd[%u]: %s\n", systemlog_host,
-        (unsigned int) (session.pid ? session.pid : getpid()), s);
+      len = pr_snprintf(buf + buflen, sizeof(buf) - buflen,
+        "%s[%u]: %s\n", process_label, (unsigned int) log_pid, s);
     }
 
     buflen += len;
@@ -486,6 +591,7 @@ static void log_write(int priority, int f, char *s, int discard) {
       buf, buflen);
 
     fprintf(stderr, "%s", buf);
+    fflush(stderr);
     return;
   }
 
@@ -497,10 +603,26 @@ static void log_write(int priority, int f, char *s, int discard) {
     }
   }
 
-  max_priority = get_param_ptr(main_server->conf, "SyslogLevel", FALSE);
-  if (max_priority != NULL &&
-      priority > *max_priority) {
+  if (main_server != NULL) {
+    ptr = get_param_ptr(main_server->conf, "SyslogLevel", FALSE);
+  }
 
+  if (ptr != NULL) {
+    max_priority = *ptr;
+
+  } else {
+    /* Default SyslogLevel is NOTICE.  Note, however, that for backward
+     * compatibility of debugging, if the DebugLevel is set higher
+     * than DEBUG0, we will automatically ASSUME that the admin wants
+     * the syslog level to be e.g. DEBUG.
+     */
+    max_priority = default_level;
+    if (debug_level != DEBUG0) {
+      max_priority = PR_LOG_DEBUG;
+    }
+  }
+
+  if (priority > max_priority) {
     /* Only return now if we don't have any log listeners. */
     if (pr_log_event_listening(PR_LOG_TYPE_SYSLOG) <= 0 &&
         pr_log_event_listening(PR_LOG_TYPE_SYSTEMLOG) <= 0) {
@@ -510,37 +632,65 @@ static void log_write(int priority, int f, char *s, int discard) {
 
   if (systemlog_fd != -1) {
     char buf[LOGBUFFER_SIZE] = {'\0'};
-    size_t buflen, len;
-    struct timeval now;
-    struct tm *tm;
-    unsigned long millis;
+    size_t buflen = 0, len = 0;
+    pid_t log_pid;
+    const char *process_label = "proftpd";
 
-    gettimeofday(&now, NULL);
-    tm = pr_localtime(NULL, (const time_t *) &(now.tv_sec));
-    if (tm == NULL) {
-      return;
+    if (log_opts & PR_LOG_OPT_USE_TIMESTAMP) {
+      pool *tmp_pool;
+      struct timeval now;
+      struct tm *tm;
+      unsigned long millis;
+
+      tmp_pool = make_sub_pool(permanent_pool);
+      pr_pool_tag(tmp_pool, "Log message pool");
+
+      gettimeofday(&now, NULL);
+      tm = pr_localtime(tmp_pool, (const time_t *) &(now.tv_sec));
+      if (tm == NULL) {
+        destroy_pool(tmp_pool);
+        return;
+      }
+
+      len = strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", tm);
+      buflen = len;
+      buf[sizeof(buf) - 1] = '\0';
+      destroy_pool(tmp_pool);
+
+      /* Convert microsecs to millisecs. */
+      millis = now.tv_usec / 1000;
+
+      len = pr_snprintf(buf + buflen, sizeof(buf) - len, ",%03lu ", millis);
+      buflen += len;
     }
 
-    len = strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", tm);
-    buflen = len;
     buf[sizeof(buf) - 1] = '\0';
 
-    /* Convert microsecs to millisecs. */
-    millis = now.tv_usec / 1000;
+    if (log_opts & PR_LOG_OPT_USE_HOSTNAME) {
+      len = pr_snprintf(buf + buflen, sizeof(buf) - buflen, "%s ",
+        systemlog_host);
+      buflen += len;
+    }
 
-    len = snprintf(buf + buflen, sizeof(buf) - len, ",%03lu ", millis);
-    buflen += len;
-    buf[sizeof(buf) - 1] = '\0';
+    log_pid = session.pid ? session.pid : getpid();
+
+    if (log_opts & PR_LOG_OPT_USE_ROLE_BASED_PROCESS_LABELS) {
+      if (is_master) {
+        process_label = "daemon";
+
+      } else {
+        process_label = "session";
+      }
+    }
 
     if (*serverinfo) {
-      len = snprintf(buf + buflen, sizeof(buf) - buflen,
-        "%s proftpd[%u] %s: %s\n", systemlog_host,
-        (unsigned int) (session.pid ? session.pid : getpid()), serverinfo, s);
+      len = pr_snprintf(buf + buflen, sizeof(buf) - buflen,
+        "%s[%u] %s: %s\n", process_label, (unsigned int) log_pid, serverinfo,
+        s);
 
     } else {
-      len = snprintf(buf + buflen, sizeof(buf) - buflen,
-        "%s proftpd[%u]: %s\n", systemlog_host,
-        (unsigned int) (session.pid ? session.pid : getpid()), s);
+      len = pr_snprintf(buf + buflen, sizeof(buf) - buflen,
+        "%s[%u]: %s\n", process_label, (unsigned int) log_pid, s);
     }
 
     buflen += len;
@@ -560,8 +710,7 @@ static void log_write(int priority, int f, char *s, int discard) {
       return;
     }
 
-    if (max_priority != NULL &&
-        priority > *max_priority) {
+    if (priority > max_priority) {
       return;
     }
 
@@ -571,7 +720,7 @@ static void log_write(int priority, int f, char *s, int discard) {
         continue;
       }
 
-      return;
+      break;
     }
 
     return;
@@ -580,11 +729,18 @@ static void log_write(int priority, int f, char *s, int discard) {
   pr_log_event_generate(PR_LOG_TYPE_SYSLOG, syslog_sockfd, priority, s,
     strlen(s));
 
-  if (set_facility != -1)
+  if (set_facility != -1) {
     f = set_facility;
+  }
 
   if (!syslog_open) {
     syslog_sockfd = pr_openlog("proftpd", LOG_NDELAY|LOG_PID, f);
+    if (syslog_sockfd < 0) {
+      (void) pr_trace_msg(trace_channel, 1,
+        "error opening syslog fd: %s", strerror(errno));
+      return;
+    }
+
     syslog_open = TRUE;
 
   } else if (f != facility) {
@@ -609,7 +765,7 @@ void pr_log_pri(int priority, const char *fmt, ...) {
   va_list msg;
 
   va_start(msg, fmt);
-  vsnprintf(buf, sizeof(buf), fmt, msg);
+  pr_vsnprintf(buf, sizeof(buf), fmt, msg);
   va_end(msg);
 
   /* Always make sure the buffer is NUL-terminated. */
@@ -626,7 +782,7 @@ void pr_log_auth(int priority, const char *fmt, ...) {
   va_list msg;
 
   va_start(msg, fmt);
-  vsnprintf(buf, sizeof(buf), fmt, msg);
+  pr_vsnprintf(buf, sizeof(buf), fmt, msg);
   va_end(msg);
 
   /* Always make sure the buffer is NUL-terminated. */
@@ -643,7 +799,7 @@ void log_stderr(int bool) {
   logstderr = bool;
 }
 
-/* Set the debug logging level, see log.h for constants.  Higher
+/* Set the debug logging level; see log.h for constants.  Higher
  * numbers mean print more, DEBUG0 (0) == print no debugging log
  * (default)
  */
@@ -653,6 +809,18 @@ int pr_log_setdebuglevel(int level) {
   return old_level;
 }
 
+/* Set the default logging level; see log.h for constants. */
+int pr_log_setdefaultlevel(int level) {
+  int old_level = default_level;
+  default_level = level;
+  return old_level;
+}
+
+int pr_log_set_options(unsigned long opts) {
+  log_opts = opts;
+  return 0;
+}
+
 /* Convert a string into the matching syslog level value.  Return -1
  * if no matching level is found.
  */
@@ -706,7 +874,7 @@ void pr_log_debug(int level, const char *fmt, ...) {
 
   memset(buf, '\0', sizeof(buf));
   va_start(msg, fmt);
-  vsnprintf(buf, sizeof(buf), fmt, msg);
+  pr_vsnprintf(buf, sizeof(buf), fmt, msg);
   va_end(msg);
 
   /* Always make sure the buffer is NUL-terminated. */
@@ -797,12 +965,83 @@ int pr_log_event_listening(unsigned int log_type) {
   return TRUE;
 }
 
+void pr_log_stacktrace(int log_fd, const char *name) {
+#if defined(HAVE_EXECINFO_H) && \
+    defined(HAVE_BACKTRACE) && \
+    defined(HAVE_BACKTRACE_SYMBOLS)
+  void *trace[PR_TUNABLE_CALLER_DEPTH];
+  int tracesz, use_fd = TRUE;
+
+  if (log_fd < 0 ||
+      name == NULL) {
+    use_fd = FALSE;
+  }
+
+  if (use_fd) {
+    (void) pr_log_writefile(log_fd, name, "%s", "-----BEGIN STACK TRACE-----");
+
+  } else {
+    (void) pr_log_pri(PR_LOG_WARNING, "-----BEGIN STACK TRACE-----");
+  }
+
+  tracesz = backtrace(trace, PR_TUNABLE_CALLER_DEPTH);
+  if (tracesz < 0) {
+    if (use_fd) {
+      (void) pr_log_writefile(log_fd, name, "backtrace(3) error: %s",
+        strerror(errno));
+
+    } else {
+      (void) pr_log_pri(PR_LOG_WARNING, "backtrace(3) error: %s",
+        strerror(errno));
+    }
+
+  } else {
+    char **strings;
+
+    strings = backtrace_symbols(trace, tracesz);
+    if (strings != NULL) {
+      register int i;
+
+      for (i = 1; i < tracesz; i++) {
+        if (use_fd) {
+          (void) pr_log_writefile(log_fd, name, "[%d] %s", i-1, strings[i]);
+
+        } else {
+          (void) pr_log_pri(PR_LOG_WARNING, "[%d] %s", i-1, strings[i]);
+        }
+      }
+
+      /* Prevent memory leaks. */
+      free(strings);
+
+    } else {
+      if (use_fd) {
+        (void) pr_log_writefile(log_fd, name,
+          "error obtaining backtrace symbols: %s", strerror(errno));
+
+      } else {
+        (void) pr_log_pri(PR_LOG_WARNING,
+          "error obtaining backtrace symbols: %s", strerror(errno));
+      }
+    }
+  }
+
+  if (use_fd) {
+    (void) pr_log_writefile(log_fd, name, "%s", "-----END STACK TRACE-----");
+
+  } else {
+    (void) pr_log_pri(PR_LOG_WARNING, "%s", "-----END STACK TRACE-----");
+  }
+#endif
+}
+
 void init_log(void) {
   char buf[256];
 
   memset(buf, '\0', sizeof(buf));
-  if (gethostname(buf, sizeof(buf)) == -1)
+  if (gethostname(buf, sizeof(buf)) < 0) {
     sstrncpy(buf, "localhost", sizeof(buf));
+  }
 
   sstrncpy(systemlog_host, (char *) pr_netaddr_validate_dns_str(buf),
     sizeof(systemlog_host));
