@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2003-2011 The ProFTPD Project team
+ * Copyright (c) 2003-2017 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,9 +22,7 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/* ProFTPD xferlog(5) logging support.
- * $Id: xferlog.c,v 1.11 2011/05/23 21:22:24 castaglia Exp $
- */
+/* ProFTPD xferlog(5) logging support. */
 
 #include "conf.h"
 
@@ -41,33 +39,38 @@ void xferlog_close(void) {
 
 int xferlog_open(const char *path) {
 
-  if (!path) {
-    if (xferlogfd != -1)
+  if (path == NULL) {
+    if (xferlogfd != -1) {
       xferlog_close();
+    }
 
     return 0;
   }
 
-  if (xferlogfd == -1) {
+  if (xferlogfd < 0) {
     pr_log_debug(DEBUG6, "opening TransferLog '%s'", path);
-    pr_log_openfile(path, &xferlogfd, PR_LOG_XFER_MODE);
+    pr_log_openfile(path, &xferlogfd, PR_TUNABLE_XFER_LOG_MODE);
 
     if (xferlogfd < 0) {
+      int xerrno = errno;
+
       pr_log_pri(PR_LOG_NOTICE, "unable to open TransferLog '%s': %s",
-        path, strerror(errno));
+        path, strerror(xerrno));
+
+      errno = xerrno;
     }
   }
 
   return xferlogfd;
 }
 
-int xferlog_write(long xfertime, const char *remhost, off_t fsize, char *fname,
-    char xfertype, char direction, char access_mode, char *user,
-    char abort_flag, const char *action_flags) {
-  const char *xfer_proto;
+int xferlog_write(long xfertime, const char *remhost, off_t fsize,
+    const char *fname, char xfertype, char direction, char access_mode,
+    const char *user, char abort_flag, const char *action_flags) {
+  pool *tmp_pool;
+  const char *rfc1413_ident = NULL, *xfer_proto;
   char buf[LOGBUFFER_SIZE] = {'\0'}, fbuf[LOGBUFFER_SIZE] = {'\0'};
   int have_ident = FALSE, len;
-  char *rfc1413_ident = NULL;
   register unsigned int i = 0;
 
   if (xferlogfd == -1 ||
@@ -78,21 +81,22 @@ int xferlog_write(long xfertime, const char *remhost, off_t fsize, char *fname,
   }
 
   for (i = 0; (i + 1 < sizeof(fbuf)) && fname[i] != '\0'; i++) {
-    fbuf[i] = (isspace((int) fname[i]) || iscntrl((int) fname[i])) ? '_' :
+    fbuf[i] = (PR_ISSPACE(fname[i]) || PR_ISCNTRL(fname[i])) ? '_' :
       fname[i];
   }
   fbuf[i] = '\0';
 
   rfc1413_ident = pr_table_get(session.notes, "mod_ident.rfc1413-ident", NULL);
-  if (rfc1413_ident) {
+  if (rfc1413_ident != NULL) {
     have_ident = TRUE;
 
     /* If the retrieved identity is "UNKNOWN", then change the string to be
      * "*", since "*" is to be logged in the xferlog, as per the doc, when
      * the authenticated user ID is not available.
      */
-    if (strncmp(rfc1413_ident, "UNKNOWN", 8) == 0)
+    if (strncmp(rfc1413_ident, "UNKNOWN", 8) == 0) {
       rfc1413_ident = "*";
+    }
 
   } else {
     /* If an authenticated user ID is not available, log "*", as per the
@@ -102,10 +106,12 @@ int xferlog_write(long xfertime, const char *remhost, off_t fsize, char *fname,
   }
 
   xfer_proto = pr_session_get_protocol(0);
+  tmp_pool = make_sub_pool(session.pool);
+  pr_pool_tag(tmp_pool, "TransferLog message pool");
 
-  len = snprintf(buf, sizeof(buf),
+  len = pr_snprintf(buf, sizeof(buf),
     "%s %ld %s %" PR_LU " %s %c %s %c %c %s %s %c %s %c\n",
-      pr_strtime(time(NULL)),
+      pr_strtime3(tmp_pool, time(NULL), FALSE),
       xfertime,
       remhost,
       (pr_off_t) fsize,
@@ -119,9 +125,10 @@ int xferlog_write(long xfertime, const char *remhost, off_t fsize, char *fname,
       have_ident ? '1' : '0',
       rfc1413_ident,
       abort_flag);
-
   buf[sizeof(buf)-1] = '\0';
 
   pr_log_event_generate(PR_LOG_TYPE_XFERLOG, xferlogfd, -1, buf, len);
+  destroy_pool(tmp_pool);
+
   return write(xferlogfd, buf, len);
 }
