@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server testsuite
- * Copyright (c) 2014-2017 The ProFTPD Project team
+ * Copyright (c) 2014-2020 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -28,7 +28,12 @@
 
 static pool *p = NULL;
 
+/* Use Google's DNS resolvers by default. */
+static const char *dns_resolver = "8.8.8.8";
+
 static void set_up(void) {
+  const char *use_resolver = NULL;
+
   if (p == NULL) {
     p = permanent_pool = make_sub_pool(NULL);
   }
@@ -37,6 +42,11 @@ static void set_up(void) {
   init_netio();
   init_inet();
 
+  use_resolver = getenv("PR_USE_DNS_RESOLVER");
+  if (use_resolver != NULL) {
+    dns_resolver = use_resolver;
+  }
+
   if (getenv("TEST_VERBOSE") != NULL) {
     pr_trace_set_levels("inet", 1, 20);
   }
@@ -508,24 +518,30 @@ START_TEST (inet_connect_ipv4_test) {
   conn = pr_inet_create_conn(p, sockfd, NULL, port, FALSE);
   fail_unless(conn != NULL, "Failed to create conn: %s", strerror(errno));
 
-  res = pr_inet_connect(p, conn, NULL, 80);
+  res = pr_inet_connect(p, conn, NULL, 180);
   fail_unless(res < 0, "Failed to handle null address");
   fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
     strerror(errno), errno);
 
+  mark_point();
   addr = pr_netaddr_get_addr(p, "127.0.0.1", NULL);
   fail_unless(addr != NULL, "Failed to resolve '127.0.0.1': %s",
     strerror(errno));
 
-  res = pr_inet_connect(p, conn, addr, 80);
-  fail_unless(res < 0, "Connected to 127.0.0.1#80 unexpectedly");
+  /* On CirrusCI VMs, attempting to connect causes test timeouts. */
+  if (getenv("CIRRUS_CLONE_DEPTH") != NULL) {
+    return;
+  }
+
+  mark_point();
+  res = pr_inet_connect(p, conn, addr, 180);
+  fail_unless(res < 0, "Connected to 127.0.0.1#180 unexpectedly");
   fail_unless(errno == ECONNREFUSED, "Expected ECONNREFUSED (%d), got %s (%d)",
     ECONNREFUSED, strerror(errno), errno);
 
-  /* Try connecting to Google's DNS server. */
-
-  addr = pr_netaddr_get_addr(p, "8.8.8.8", NULL);
-  fail_unless(addr != NULL, "Failed to resolve '8.8.8.8': %s",
+#if defined(PR_USE_NETWORK_TESTS)
+  addr = pr_netaddr_get_addr(p, dns_resolver, NULL);
+  fail_unless(addr != NULL, "Failed to resolve '%s': %s", dns_resolver,
     strerror(errno));
 
   res = pr_inet_connect(p, conn, addr, 53);
@@ -543,14 +559,16 @@ START_TEST (inet_connect_ipv4_test) {
   fail_unless(conn != NULL, "Failed to create conn: %s", strerror(errno));
 
   res = pr_inet_connect(p, conn, addr, 53);
-  fail_if(res < 0, "Failed to connect to 8.8.8.8#53: %s", strerror(errno));
+  fail_if(res < 0, "Failed to connect to %s#53: %s", dns_resolver,
+    strerror(errno));
 
   res = pr_inet_connect(p, conn, addr, 53);
-  fail_unless(res < 0, "Failed to connect to 8.8.8.8#53: %s",
+  fail_unless(res < 0, "Failed to connect to %s#53: %s", dns_resolver,
     strerror(errno));
   fail_unless(errno == EISCONN, "Expected EISCONN (%d), got %s (%d)",
     EISCONN, strerror(errno), errno);
   pr_inet_close(p, conn);
+#endif
 }
 END_TEST
 
@@ -569,16 +587,24 @@ START_TEST (inet_connect_ipv6_test) {
   conn = pr_inet_create_conn(p, sockfd, NULL, port, FALSE);
   fail_unless(conn != NULL, "Failed to create conn: %s", strerror(errno));
 
+  mark_point();
   addr = pr_netaddr_get_addr(p, "::1", NULL);
   fail_unless(addr != NULL, "Failed to resolve '::1': %s",
     strerror(errno));
 
-  res = pr_inet_connect(p, conn, addr, 80);
-  fail_unless(res < 0, "Connected to ::1#80 unexpectedly");
+  /* On CirrusCI VMs, attempting to connect causes test timeouts. */
+  if (getenv("CIRRUS_CLONE_DEPTH") != NULL) {
+    return;
+  }
+
+  mark_point();
+  res = pr_inet_connect(p, conn, addr, 180);
+  fail_unless(res < 0, "Connected to ::1#180 unexpectedly");
   fail_unless(errno == ECONNREFUSED || errno == ENETUNREACH || errno == EADDRNOTAVAIL,
     "Expected ECONNREFUSED (%d), ENETUNREACH (%d), or EADDRNOTAVAIL (%d), got %s (%d)",
     ECONNREFUSED, ENETUNREACH, EADDRNOTAVAIL, strerror(errno), errno);
 
+#if defined(PR_USE_NETWORK_TESTS)
   /* Try connecting to Google's DNS server. */
 
   addr = pr_netaddr_get_addr(p, "2001:4860:4860::8888", NULL);
@@ -614,6 +640,7 @@ START_TEST (inet_connect_ipv6_test) {
   fail_unless(errno == EISCONN || errno == EHOSTUNREACH || errno == ENETUNREACH || errno == EADDRNOTAVAIL,
     "Expected EISCONN (%d) or EHOSTUNREACH (%d) or ENETUNREACH (%d) or EADDRNOTAVAIL (%d), got %s (%d)", EISCONN, EHOSTUNREACH, ENETUNREACH, EADDRNOTAVAIL, strerror(errno), errno);
   pr_inet_close(p, conn);
+#endif
 
   pr_inet_set_default_family(p, AF_INET);
 
@@ -637,7 +664,7 @@ START_TEST (inet_connect_nowait_test) {
   conn = pr_inet_create_conn(p, sockfd, NULL, port, FALSE);
   fail_unless(conn != NULL, "Failed to create conn: %s", strerror(errno));
 
-  res = pr_inet_connect_nowait(p, conn, NULL, 80);
+  res = pr_inet_connect_nowait(p, conn, NULL, 180);
   fail_unless(res < 0, "Failed to handle null address");
   fail_unless(errno == EINVAL, "Expected EINVAL (%d), got %s (%d)", EINVAL,
     strerror(errno), errno);
@@ -646,23 +673,26 @@ START_TEST (inet_connect_nowait_test) {
   fail_unless(addr != NULL, "Failed to resolve '127.0.0.1': %s",
     strerror(errno));
 
-  res = pr_inet_connect_nowait(p, conn, addr, 80);
-  fail_unless(res != -1, "Connected to 127.0.0.1#80 unexpectedly");
+  res = pr_inet_connect_nowait(p, conn, addr, 180);
+  fail_unless(res != -1, "Connected to 127.0.0.1#180 unexpectedly");
 
+#if defined(PR_USE_NETWORK_TESTS)
   /* Try connecting to Google's DNS server. */
 
-  addr = pr_netaddr_get_addr(p, "8.8.8.8", NULL);
-  fail_unless(addr != NULL, "Failed to resolve '8.8.8.8': %s",
+  addr = pr_netaddr_get_addr(p, dns_resolver, NULL);
+  fail_unless(addr != NULL, "Failed to resolve '%s': %s", dns_resolver,
     strerror(errno));
 
   res = pr_inet_connect_nowait(p, conn, addr, 53);
   if (res < 0 &&
-      errno != ECONNREFUSED) {
-    fail_unless(res != -1, "Failed to connect to 8.8.8.8#53: %s",
+      errno != ECONNREFUSED &&
+      errno != EBADF) {
+    fail_unless(res != -1, "Failed to connect to %s#53: %s", dns_resolver,
       strerror(errno));
   }
 
   pr_inet_close(p, conn);
+#endif
 
   /* Restore the default family to AF_INET, for other tests. */
   pr_inet_set_default_family(p, AF_INET);
