@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001-2010 The ProFTPD Project team
+ * Copyright (c) 2001-2016 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project and other respective copyright
  * holders give permission to link this program with OpenSSL, and distribute
@@ -22,9 +22,7 @@
  * the source distribution.
  */
 
-/* "Scrubs" the scoreboard file, clearing it of old/stale entries.
- * $Id: ftpscrub.c,v 1.1.2.1 2010/03/25 17:37:49 castaglia Exp $
- */
+/* "Scrubs" the scoreboard file, clearing it of old/stale entries. */
 
 #include "utils.h"
 
@@ -33,8 +31,9 @@ static const char *config_filename = PR_CONFIG_FILE_PATH;
 static int check_scoreboard_file(void) {
   struct stat st;
 
-  if (stat(util_get_scoreboard(), &st) < 0)
+  if (stat(util_get_scoreboard(), &st) < 0) {
     return -1;
+  }
 
   return 0;
 }
@@ -69,10 +68,12 @@ static void show_usage(const char *progname, int exit_code) {
 #else /* HAVE_GETOPT_LONG */
     printf("  %s\n", h->short_opt);
 #endif
-    if (!h->desc)
+    if (h->desc == NULL) {
       printf("    display %s usage\n", progname);
-    else
+
+    } else {
       printf("    %s\n", h->desc);
+    }
   }
 
   exit(exit_code);
@@ -99,6 +100,7 @@ int main(int argc, char **argv) {
     switch (c) {
       case 'h':
         show_usage(progname, 0);
+        break;
 
       case 'f':
         util_set_scoreboard(optarg);
@@ -115,6 +117,7 @@ int main(int argc, char **argv) {
       case '?':
         fprintf(stderr, "unknown option: %c\n", (char) optopt);
         show_usage(progname, 1);
+        break;
     }
   }
 
